from typing import Any, List, Optional, Union
import torch
from pydantic import BaseModel, field_validator

class DatasetParams(BaseModel):
    blur_gaussian_sigma: Optional[List[float]]
    stability_threshold: Union[List[float], float]
    correctness_params: List[float]
    downsample_method: str
    upsample_method: str


class Metric(BaseModel):
    value: Any
    description: str

    @classmethod
    @field_validator("value")
    def check_value(cls, value) -> torch.Tensor:
        if not isinstance(value, torch.Tensor):
            raise ValueError("value must be a torch.Tensor.")
        return value


class Consistency(BaseModel):
    reflectance: Metric
    spectral: Metric
    spatial: Metric


class Distance(BaseModel):
    lr_to_hr: Any
    sr_to_hr: Any
    sr_to_lr: Any

    @classmethod
    @field_validator("lr_to_hr", "sr_to_hr", "sr_to_lr")
    def check_value(cls, value) -> torch.Tensor:
        if not isinstance(value, torch.Tensor):
            raise ValueError("value must be a torch.Tensor.")
        return value


class Correctness(BaseModel):
    omission: Any
    improvement: Any
    hallucination: Any
    classification: Any

    @classmethod
    @field_validator("omission", "improvement", "hallucination", "classification")
    def check_value(cls, value) -> torch.Tensor:
        if not isinstance(value, torch.Tensor):
            raise ValueError("value must be a torch.Tensor.")
        return value


class Auxiliar(BaseModel):
    sr_harm: Any
    lr_to_hr: Any
    matching_points_lr: Union[dict, bool]
    matching_points_hr: Union[dict, bool]

    @classmethod
    @field_validator("sr_harm", "lr_to_hr")
    def check_value(cls, value) -> torch.Tensor:
        if not isinstance(value, torch.Tensor):
            raise ValueError("value must be a torch.Tensor.")
        return value


class Results(BaseModel):
    consistency: Consistency
    distance: Distance
    synthesis: Any
    correctness: Correctness
    auxiliar: Auxiliar


class Config(BaseModel):
    # General parameters
    agg_method: str = "pixel"  # pixel, image
    mask: Optional[int] = 32
    patch_size: Optional[int] = None        
    rgb_bands: Optional[List[int]] = [0, 1, 2]

    # Spatial parameters
    spatial_features: str = "superpoint"
    spatial_matcher: str = "lightglue"
    spatial_max_num_keypoints: int = 1000
    spatial_threshold_distance: int = 3
    spatial_threshold_npoints: int = 5

    # Spectral parameters
    reflectance_method: str = "l1"
    spectral_method: str = "sad"

    # Create SRharm
    harm_apply_spectral: bool = True
    harm_apply_spatial: bool = True

    # Unsystematic error parameters
    distance_method: str = "l1"

    # General parameters - validator ----------------------------
    @field_validator("agg_method")
    @classmethod
    def check_agg_method(cls, value) -> str:
        valid_methods = ["pixel", "image", "patch"]
        if value not in valid_methods:
            raise ValueError(f"Invalid method. Must be one of {valid_methods}")
        return value

    @field_validator("rgb_bands")
    @classmethod
    def check_rgb_bands(cls, value) -> List[int]:
        if len(value) != 3:
            raise ValueError("rgb_bands must have 3 elements.")
        return value


    # Spatial parameters - validator ----------------------------

    @field_validator("spatial_features")
    @classmethod
    def check_spatial_features(cls, value) -> str:
        valid_methods = ["disk", "superpoint"]
        if value not in valid_methods:
            raise ValueError(
                f"Invalid spatial_features. Must be one of {valid_methods}"
            )
        return value

    @field_validator("spatial_matcher")
    @classmethod
    def check_spatial_matcher(cls, value) -> str:
        valid_methods = ["lightglue", "superglue"]
        if value not in valid_methods:
            raise ValueError(f"Invalid spatial_matcher. Must be one of {valid_methods}")
        return value

    @field_validator("spatial_max_num_keypoints")
    @classmethod
    def check_spatial_max_num_keypoints(cls, value) -> int:
        if value < 0:
            raise ValueError("spatial_max_num_keypoints must be positive.")
        return value

    @field_validator("spatial_threshold_distance")
    @classmethod
    def check_spatial_threshold_distance(cls, value) -> int:
        if value < 0:
            raise ValueError("spatial_threshold_distance must be positive.")
        return value

    @field_validator("spatial_threshold_npoints")
    @classmethod
    def check_spatial_threshold_npoints(cls, value) -> int:
        if value < 0:
            raise ValueError("spatial_threshold_npoints must be positive.")
        return value

    # Spectral parameters - validator ----------------------------

    # Global ---------------------------------
    @field_validator("reflectance_method")
    @classmethod
    def check_reflectance_method(cls, value) -> str:
        valid_methods = ["kl", "l1", "l2", "pbias"]
        if value not in valid_methods:
            raise ValueError(
                f"Invalid reflectance_method. Must be one of {valid_methods}"
            )
        return value

    # Local ---------------------------------
    @field_validator("spectral_method")
    @classmethod
    def check_spectral_method(cls, value) -> str:
        valid_methods = ["sad"]
        if value not in valid_methods:
            raise ValueError(f"Invalid spectral_method. Must be one of {valid_methods}")
        return value

    # Create SRharm - validator ------------------------------------------
    @field_validator("harm_apply_spectral")
    @classmethod
    def check_harm_apply_spectral(cls, value) -> bool:
        if not isinstance(value, bool):
            raise ValueError("harm_apply_spectral must be boolean.")
        return value

    @field_validator("harm_apply_spatial")
    @classmethod
    def check_harm_apply_spatial(cls, value) -> bool:
        if not isinstance(value, bool):
            raise ValueError("harm_apply_spatial must be boolean.")
        return value

    # Create Unsystematic error parameters - validator -------------------------

    @field_validator("distance_method")
    @classmethod
    def check_unsys_method(cls, value) -> str:
        valid_methods = ["psnr", "lpips", "clip", "sam", "l1", "l2", "kl", "pbias"]
        if value not in valid_methods:
            raise ValueError(f"Invalid unsys_method. Must be one of {valid_methods}")
        return value

def create_param_config(dataset = "naip") -> DatasetParams:
    if dataset == "naip":
        params = DatasetParams(
            blur_gaussian_sigma = [0.5291, 0.4943, 0.5110, 0.4771],
            stability_threshold = [
                0.02, 0.016, 0.025, 0.015, 0.023, 0.019,
                0.021, 0.019, 0.012, 0.027, 0.016, 0.02,
                0.018, 0.02, 0.018, 0.019, 0.019, 0.019,
                0.013, 0.013, 0.013, 0.015, 0.015, 0.015,
                0.018, 0.017, 0.016, 0.015, 0.022, 0.016
            ],
            correctness_params=[0.80, 0.80, 0.40],
            downsample_method = "naip",
            upsample_method = "classic"
        )
    elif dataset == "spot":
        params = DatasetParams(
            blur_gaussian_sigma = [0.5795, 0.6057, 0.6451, 0.6145],
            stability_threshold = [
                0.014, 0.016, 0.013, 0.010, 0.032,
                0.040, 0.014, 0.037, 0.030, 0.036,
                0.026, 0.030
            ],
            correctness_params=[0.80, 0.80, 0.40],
            downsample_method = "spot",
            upsample_method = "classic"
        )

    elif dataset == "venus":
        params = DatasetParams(
            blur_gaussian_sigma = [0.5180, 0.5305, 0.5468, 0.5645],
            stability_threshold = [
                0.015, 0.012, 0.012, 0.012, 0.013, 0.011,
                0.01, 0.012, 0.011, 0.013, 0.01, 0.013,
                0.013, 0.01, 0.017, 0.01, 0.012, 0.013,
                0.011, 0.013, 0.017, 0.011, 0.01, 0.011,
                0.012, 0.015, 0.012, 0.011, 0.01, 0.017,
                0.016, 0.01, 0.011, 0.014, 0.012, 0.016,
                0.01, 0.012, 0.013, 0.01, 0.016, 0.016, 
                0.012, 0.011, 0.016, 0.01, 0.015, 0.016, 
                0.016, 0.01, 0.015, 0.011, 0.016, 0.016, 
                0.013, 0.013, 0.014, 0.01, 0.011
            ],
            correctness_params=[0.80, 0.80, 0.40],
            downsample_method = "venus",
            upsample_method = "classic"
        )
    elif dataset == "general":
        params = DatasetParams(
            blur_gaussian_sigma = None,
            stability_threshold = 0.01,
            correctness_params=[0.80, 0.80, 0.40],
            downsample_method = "classic",
            upsample_method = "classic"
        )

    return params


import torch
from typing import List, Optional


def classic_upsampling(x: torch.Tensor, scale: int) -> torch.Tensor:
    """ Upsampling a tensor (B, C, H, W) to a lower resolution 
    (B, C, H', W') using bilinear interpolation with antialiasing.

    Args:
        x (torch.Tensor): The tensor to upsample.
        scale (int, optional): The super-resolution scale. Defaults 
            to 4.

    Returns:
        torch.Tensor: The upsampled tensor (B, C, H', W').
    """

    x_ref = torch.nn.functional.interpolate(
        input=x, scale_factor=1 / scale, mode="bilinear", antialias=True
    )

    return x_ref


def naip_upsampling(
        x: torch.Tensor,
        params: Optional[List[float]] = [0.5291, 0.4943, 0.5110, 0.4771],
        rgb_bands: Optional[List[int]] = [0, 1, 2]
) -> torch.Tensor:
    """ Upsampling a tensor (B, C, H, W) to a lower resolution 
    (B, C, H', W') using a gaussian blur and upsampling
    withouth antialiasing. The blur kernel is trained using
    the curated dataset opensr-test-spot.

    Args:
        x (torch.Tensor): The tensor to upsample.
        scale (int, optional): The super-resolution scale. Defaults 
            to 4.

    Returns:
        torch.Tensor: The upsampled tensor (B, C, H', W').
    """
    # get the nir band according to the rgb bands    
    if x.shape[0] == 4:
        nir_band_index = [i for i in range(4) if i not in rgb_bands][0]
        x = x[rgb_bands + [nir_band_index], ...]
    elif x.shape[0] == 3:
        x = x[rgb_bands, ...]
    else:
        raise ValueError("The input tensor must have 3 or 4 bands.")

    # Blur 5x5 kernel by band trained using the
    # curated dataset opensr-test-naip
    with torch.no_grad():
        blur_R = GaussianBlur(kernel_size=7, params=[params[0]])
        blur_G = GaussianBlur(kernel_size=7, params=[params[1]])
        blur_B = GaussianBlur(kernel_size=7, params=[params[2]])
        blur_N = GaussianBlur(kernel_size=7, params=[params[3]])

        # Apply the blur kernel to each band
        container = []
        for i in range(x.shape[0]):
            if i == 0:
                _x_blurred = blur_R(x[i][None, None, ...]).squeeze()
            elif i == 1:
                _x_blurred = blur_G(x[i][None, None, ...]).squeeze()
            elif i == 2:
                _x_blurred = blur_B(x[i][None, None, ...]).squeeze()
            elif i == 3:
                _x_blurred = blur_N(x[i][None, None, ...]).squeeze()
            container.append(_x_blurred)
        
    # Apply the blur kernel to each band
    x_blurred = torch.stack(container, dim=0)
    
    # Downsample using bilinear interpolation
    x_ref = torch.nn.functional.interpolate(
        input=x_blurred[None], scale_factor=0.25, mode="bilinear", antialias=False
    ).squeeze()

    return x_ref


def spot_upsampling(
    x: torch.Tensor,
    rgb_bands: Optional[List[int]] = [0, 1, 2],
    params: Optional[List[float]] = [0.5795, 0.6057, 0.6451, 0.6145]
) -> torch.Tensor:
    """ Upsampling a tensor (B, C, H, W) to a lower resolution 
    (B, C, H', W') using a gaussian blur and upsampling
    withouth antialiasing. The blur kernel is trained using
    the curated dataset opensr-test-spot.

    Args:
        x (torch.Tensor): The tensor to upsample.
        scale (int, optional): The super-resolution scale. Defaults 
            to 4.

    Returns:
        torch.Tensor: The upsampled tensor (B, C, H', W').
    """
    # get the nir band according to the rgb bands    
    if x.shape[0] == 4:
        nir_band_index = [i for i in range(4) if i not in rgb_bands][0]
        x = x[rgb_bands + [nir_band_index], ...]
    elif x.shape[0] == 3:
        x = x[rgb_bands, ...]
    else:
        raise ValueError("The input tensor must have 3 or 4 bands.")

    # Blur 5x5 kernel by band trained using the
    # curated dataset opensr-test-naip
    with torch.no_grad():
        blur_R = GaussianBlur(kernel_size=7, params=[params[0]])
        blur_G = GaussianBlur(kernel_size=7, params=[params[1]])
        blur_B = GaussianBlur(kernel_size=7, params=[params[2]])
        blur_N = GaussianBlur(kernel_size=7, params=[params[3]])

        # Apply the blur kernel to each band
        container = []
        for i in range(x.shape[0]):
            if i == 0:
                _x_blurred = blur_R(x[i][None, None, ...]).squeeze()
            elif i == 1:
                _x_blurred = blur_G(x[i][None, None, ...]).squeeze()
            elif i == 2:
                _x_blurred = blur_B(x[i][None, None, ...]).squeeze()
            elif i == 3:
                _x_blurred = blur_N(x[i][None, None, ...]).squeeze()
            container.append(_x_blurred)
        
    # Apply the blur kernel to each band
    x_blurred = torch.stack(container, dim=0)
    
    # Downsample using bilinear interpolation
    x_ref = torch.nn.functional.interpolate(
        input=x_blurred[None], scale_factor=0.25, mode="bilinear", antialias=False
    ).squeeze()

    return x_ref


def venus_upsampling(
    x: torch.Tensor,
    params: Optional[List[float]] = [0.5180, 0.5305, 0.5468, 0.5645],
    rgb_bands: Optional[List[int]] = [0, 1, 2]
) -> torch.Tensor:
    """ Upsampling a tensor (B, C, H, W) to a lower resolution
    (B, C, H', W') using a gaussian blur and upsampling
    withouth antialiasing. The blur kernel is trained using
    the curated dataset opensr-test-venus.

    Args:
        x (torch.Tensor): The tensor to upsample.
        scale (int, optional): The super-resolution scale. Defaults
            to 4.

    Returns:
        torch.Tensor: The upsampled tensor (B, C, H', W').
    """
    # get the nir band according to the rgb bands    
    if x.shape[0] == 4:
        nir_band_index = [i for i in range(4) if i not in rgb_bands][0]
        x = x[rgb_bands + [nir_band_index], ...]
    elif x.shape[0] == 3:
        x = x[rgb_bands, ...]
    else:
        raise ValueError("The input tensor must have 3 or 4 bands.")

    # Blur 5x5 kernel by band trained using the
    # curated dataset opensr-test-naip
    with torch.no_grad():
        blur_R = GaussianBlur(kernel_size=3, params=[params[0]])
        blur_G = GaussianBlur(kernel_size=3, params=[params[1]])
        blur_B = GaussianBlur(kernel_size=3, params=[params[2]])
        blur_N = GaussianBlur(kernel_size=3, params=[params[3]])

        # Apply the blur kernel to each band
        container = []
        for i in range(x.shape[0]):
            if i == 0:
                _x_blurred = blur_R(x[i][None, None, ...]).squeeze()
            elif i == 1:
                _x_blurred = blur_G(x[i][None, None, ...]).squeeze()
            elif i == 2:
                _x_blurred = blur_B(x[i][None, None, ...]).squeeze()
            elif i == 3:
                _x_blurred = blur_N(x[i][None, None, ...]).squeeze()
            container.append(_x_blurred)
        
    # Apply the blur kernel to each band
    x_blurred = torch.stack(container, dim=0)
    
    # Downsample using bilinear interpolation
    x_ref = torch.nn.functional.interpolate(
        input=x_blurred[None], scale_factor=0.5, mode="bilinear", antialias=False
    ).squeeze()

    return x_ref


def apply_upsampling(
    X: torch.Tensor, scale: int = 4, method: str = "classic"
) -> torch.Tensor:
    """ Apply a upsampling method to a tensor (B, C, H, W).

    Args:
        X (torch.Tensor): The tensor to downsample.
        scale (int, optional): The super-resolution scale. Defaults
            to 4.
        method (str, optional): The upsampling method. Must be one 
            of "classic", "naip", "spot", "venus". Defaults to "classic".

    Raises:
        ValueError: If the method is not valid.

    Returns:
        torch.Tensor: The upscaled tensor.
    """

    if method == "classic":
        return classic_upsampling(X, scale)
    elif method == "naip":
        return naip_upsampling(X)
    elif method == "spot":
        return spot_upsampling(X)
    elif method == "venus":
        return venus_upsampling(X)
    else:
        raise ValueError(
            "Invalid method. Must be one of 'classic', 'naip', 'spot', 'venus'."
        )


def classic_downsampling(x: torch.Tensor, scale: int = 4) -> torch.Tensor:
    """ Downsampling a tensor (B, C, H, W) to a upper resolution 
    (B, C, H', W') using bilinear interpolation with antialiasing.

    Args:
        x (torch.Tensor): The tensor to downsampling.
        scale (int, optional): The super-resolution scale. Defaults 
            to 4.

    Returns:
        torch.Tensor: The downscaled tensor (B, C, H', W').
    """

    x_ref = torch.nn.functional.interpolate(
        input=x, scale_factor=scale, mode="bilinear", antialias=True
    )

    return x_ref


def apply_downsampling(
    X: torch.Tensor, scale: int = 4, method: str = "classic"
) -> torch.Tensor:
    """ Apply a downsampling method to a tensor (B, C, H, W).

    Args:
        X (torch.Tensor): The tensor to downsample.
        scale (int, optional): The super-resolution scale. Defaults
            to 4.
        method (str, optional): The downsampling method. Must be one 
            of ["classic"]. Defaults to "classic".

    Raises:
        ValueError: If the method is not valid.

    Returns:
        torch.Tensor: The downscaled tensor.
    """

    if method == "classic":
        return classic_downsampling(X, scale)
    else:
        raise ValueError("Invalid method. Must be one of ['classic']")



def gaussian(x, sigma: float= 2.5):
    """
    Returns a 1D Gaussian distribution tensor.

    Args:
        x (Tensor): input tensor.
        sigma (float, optional): standard deviation of the Gaussian distribution. Default is 2.5.

    Returns:
        Tensor: 1D Gaussian distribution tensor.
    """    
    gauss = torch.exp(-x.pow(2.0) / (2 * sigma.pow(2.0)))
    return gauss / gauss.sum(-1, keepdim=True)


def gaussian2D(sigma: torch.Tensor, window_size: int = 33):
    """
    Returns a 2D Gaussian distribution tensor.

    Args:
        sigma (Tensor): standard deviation of the Gaussian distribution.
        window_size (int, optional): size of the window to apply the Gaussian
            distribution. Default is 33.
        batch_size (int, optional): size of the batch. Default is 1.

    Returns:
        Tensor: 2D Gaussian distribution tensor.
    """        
    # kernel 1D
    ky = gaussian(torch.linspace(-1, 1, window_size), sigma=sigma)
    kx = gaussian(torch.linspace(-1, 1, window_size), sigma=sigma)
    
    kernel = kx.unsqueeze(1) * ky.unsqueeze(0)
    
    return kernel

class GaussianBlur(torch.nn.Module):
    def __init__(self, kernel_size=65, params=[2.5], device=torch.device("cpu")):
        super(GaussianBlur, self).__init__()
        self.kernel_size = kernel_size
        self.sigma = torch.nn.Parameter(torch.tensor(params[0]))
        # if type is str
        if isinstance(device, str):
            self.device = torch.device(device)
        else:
            self.device = device

    def forward(self, x):
        kernel = gaussian2D(self.sigma, self.kernel_size).expand(x.shape[0], 1, self.kernel_size, self.kernel_size)
        kernel = kernel.to(self.device)
        return torch.nn.functional.conv2d(x, kernel, groups=x.shape[0], padding=self.kernel_size//2)



"""
Adapted from  https://github.com/SaoYan/DnCNN-PyTorch SaoYan
Re-implemented by Yuqian Zhou
"""
import math

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from skimage.metrics import peak_signal_noise_ratio as compare_psnr
from torch.autograd import Variable


class RealNoisyEstimator(nn.Module):
    """
    Noise estimator, with original 3 layers
    """

    def __init__(self, input_channels=1, output_channels=3, num_of_layers=3):
        super(RealNoisyEstimator, self).__init__()
        kernel_size = 3
        padding = 1
        features = 64
        layers = []
        layers.append(
            nn.Conv2d(
                in_channels=input_channels,
                out_channels=features,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        layers.append(nn.ReLU(inplace=True))
        for _ in range(num_of_layers - 2):
            layers.append(
                nn.Conv2d(
                    in_channels=features,
                    out_channels=features,
                    kernel_size=kernel_size,
                    padding=padding,
                    bias=False,
                )
            )
            layers.append(nn.BatchNorm2d(features))
            layers.append(nn.ReLU(inplace=True))
        layers.append(
            nn.Conv2d(
                in_channels=features,
                out_channels=output_channels,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        self.dncnn = nn.Sequential(*layers)

    def forward(self, input):
        x = self.dncnn(input)
        return x


class NoisyEstimator(nn.Module):
    def __init__(self, channels, num_of_layers=17, num_of_est=3):
        super(NoisyEstimator, self).__init__()
        kernel_size = 3
        padding = 1
        features = 64
        layers = []
        layers.append(
            nn.Conv2d(
                in_channels=channels + num_of_est,
                out_channels=features,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        layers.append(nn.ReLU(inplace=True))
        for _ in range(num_of_layers - 2):
            layers.append(
                nn.Conv2d(
                    in_channels=features,
                    out_channels=features,
                    kernel_size=kernel_size,
                    padding=padding,
                    bias=False,
                )
            )
            layers.append(nn.BatchNorm2d(features))
            layers.append(nn.ReLU(inplace=True))
        layers.append(
            nn.Conv2d(
                in_channels=features,
                out_channels=channels,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        self.dncnn = nn.Sequential(*layers)

    def forward(self, x, c):
        input_x = torch.cat([x, c], dim=1)
        out = self.dncnn(input_x)
        return out


def weights_init_kaiming(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        nn.init.kaiming_normal(m.weight.data, a=0, mode="fan_in")
    elif classname.find("Linear") != -1:
        nn.init.kaiming_normal(m.weight.data, a=0, mode="fan_in")
    elif classname.find("BatchNorm") != -1:
        # nn.init.uniform(m.weight.data, 1.0, 0.02)
        m.weight.data.normal_(mean=0, std=math.sqrt(2.0 / 9.0 / 64.0)).clamp_(
            -0.025, 0.025
        )
        nn.init.constant(m.bias.data, 0.0)


def batch_PSNR(img, imclean, data_range):
    Img = img.data.cpu().numpy().astype(np.float32)
    Iclean = imclean.data.cpu().numpy().astype(np.float32)
    PSNR = 0
    for i in range(Img.shape[0]):
        PSNR += compare_psnr(Iclean[i, :, :, :], Img[i, :, :, :], data_range=data_range)
    return PSNR / Img.shape[0]


def data_augmentation(image, mode):
    out = np.transpose(image, (1, 2, 0))
    if mode == 0:
        # original
        out = out
    elif mode == 1:
        # flip up and down
        out = np.flipud(out)
    elif mode == 2:
        # rotate counterwise 90 degree
        out = np.rot90(out)
    elif mode == 3:
        # rotate 90 degree and flip up and down
        out = np.rot90(out)
        out = np.flipud(out)
    elif mode == 4:
        # rotate 180 degree
        out = np.rot90(out, k=2)
    elif mode == 5:
        # rotate 180 degree and flip
        out = np.rot90(out, k=2)
        out = np.flipud(out)
    elif mode == 6:
        # rotate 270 degree
        out = np.rot90(out, k=3)
    elif mode == 7:
        # rotate 270 degree and flip
        out = np.rot90(out, k=3)
        out = np.flipud(out)
    return np.transpose(out, (2, 0, 1))


def visual_va2np(
    Out,
    mode=1,
    ps=0,
    pss=1,
    scal=1,
    rescale=0,
    w=10,
    h=10,
    c=3,
    refill=0,
    refill_img=0,
    refill_ind=[0, 0],
):
    if mode == 0 or mode == 1 or mode == 3:
        out_numpy = Out.data.squeeze(0).cpu().numpy()
    elif mode == 2:
        out_numpy = Out.data.squeeze(1).cpu().numpy()
    if out_numpy.shape[0] == 1:
        out_numpy = np.tile(out_numpy, (3, 1, 1))
    if mode == 0 or mode == 1:
        out_numpy = (np.transpose(out_numpy, (1, 2, 0))) * 255.0 * scal
    else:
        out_numpy = np.transpose(out_numpy, (1, 2, 0))

    if ps == 1:
        out_numpy = reverse_pixelshuffle(out_numpy, pss, refill, refill_img, refill_ind)
    if rescale == 1:
        out_numpy = cv2.resize(out_numpy, (h, w))
        # print(out_numpy.shape)
    return out_numpy


def temp_ps_4comb(Out, In):
    pass


def np2ts(x, mode=0):  # now assume the input only has one channel which is ignored
    w, h, c = x.shape
    x_ts = x.transpose(2, 0, 1)
    x_ts = torch.from_numpy(x_ts).type(torch.FloatTensor)
    if mode == 0 or mode == 1:
        x_ts = x_ts.unsqueeze(0)
    elif mode == 2:
        x_ts = x_ts.unsqueeze(1)
    return x_ts


def np2ts_4d(x):
    x_ts = x.transpose(0, 3, 1, 2)
    x_ts = torch.from_numpy(x_ts).type(torch.FloatTensor)
    return x_ts


def get_salient_noise_in_maps(lm, thre=0.0, chn=3):
    """
    Description: To find out the most frequent estimated noise level in the images
    ----------
    [Input]
    a multi-channel tensor of noise map

    [Output]
    A list of  noise level value
    """
    lm_numpy = lm.data.cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (0, 2, 3, 1))
    nl_list = np.zeros((lm_numpy.shape[0], chn, 1))
    for n in range(lm_numpy.shape[0]):
        for c in range(chn):
            selected_lm = np.reshape(
                lm_numpy[n, :, :, c], (lm_numpy.shape[1] * lm_numpy.shape[2], 1)
            )
            selected_lm = selected_lm[selected_lm > thre]
            if selected_lm.shape[0] == 0:
                nl_list[n, c] = 0
            else:
                hist = np.histogram(selected_lm, density=True)
                nl_ind = np.argmax(hist[0])
                # print(nl_ind)
                # print(hist[0])
                # print(hist[1])
                nl = (hist[1][nl_ind] + hist[1][nl_ind + 1]) / 2.0
                nl_list[n, c] = nl
    return nl_list


def get_cdf_noise_in_maps(lm, thre=0.8, chn=3):
    """
    Description: To find out the most frequent estimated noise level in the images
    ----------
    [Input]
    a multi-channel tensor of noise map

    [Output]
    A list of  noise level value
    """
    lm_numpy = lm.data.cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (0, 2, 3, 1))
    nl_list = np.zeros((lm_numpy.shape[0], chn, 1))
    for n in range(lm_numpy.shape[0]):
        for c in range(chn):
            selected_lm = np.reshape(
                lm_numpy[n, :, :, c], (lm_numpy.shape[1] * lm_numpy.shape[2], 1)
            )
            H, x = np.histogram(selected_lm, normed=True)
            dx = x[1] - x[0]
            F = np.cumsum(H) * dx
            F_ind = np.where(F > 0.9)[0][0]
            nl_list[n, c] = x[F_ind]
            print(nl_list[n, c])
    return nl_list


def get_pdf_in_maps(lm, mark, chn=1):
    """
    Description: get the noise estimation cdf of each channel
    ----------
    [Input]
    a multi-channel tensor of noise map and channel dimension
    chn: the channel number for gaussian
    [Output]
    CDF function of each sample and each channel
    """
    lm_numpy = lm.data.cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (0, 2, 3, 1))
    pdf_list = np.zeros((lm_numpy.shape[0], chn, 10))
    for n in range(lm_numpy.shape[0]):
        for c in range(chn):
            selected_lm = np.reshape(
                lm_numpy[n, :, :, c], (lm_numpy.shape[1] * lm_numpy.shape[2], 1)
            )
            H, x = np.histogram(selected_lm, range=(0.0, 1.0), bins=10, normed=True)
            dx = x[1] - x[0]
            F = H * dx
            pdf_list[n, c, :] = F
            # sio.savemat(mark + str(c) + '.mat',{'F':F})
            plt.bar(range(10), F)
            # plt.savefig(mark + str(c) + '.png')
            plt.close()
    return pdf_list


def get_pdf_matching_score(F1, F2):
    """
    Description: Given two sets of CDF, get the overall matching score for each channel
    -----------
    [Input] F1, F2
    [Output] score for each channel
    """
    return np.mean((F1 - F2) ** 2)


def decide_scale_factor(
    noisy_image, estimation_model, color=1, thre=0, stopping=4, mark=""
):
    """
    Description: Given a noisy image and the noise estimation model, keep multiscaling the image\\
                 using pixel-shuffle methods, and estimate the pdf and cdf of AWGN channel
                 Compare the changes of the density function and decide the optimal scaling factor
    ------------
    [Input]  noisy_image, estimation_model, plot_flag, stopping
    [Output]  plot the middle vector
              score_seq: the matching score sequence between the two subsequent pdf
              opt_scale: the optimal scaling factor 
    """
    if color == 1:
        c = 3
    elif color == 0:
        c = 1
    score_seq = []
    Pre_CDF = None
    flag = 0
    for pss in range(1, stopping + 1):  # scaling factor from 1 to the limit
        noisy_image = pixelshuffle(noisy_image, pss)
        INoisy = np2ts(noisy_image, color)
        INoisy = Variable(INoisy, volatile=True)
        EMap = torch.clamp(estimation_model(INoisy), 0.0, 1.0)
        EPDF = get_pdf_in_maps(EMap, mark + str(pss), c)[0]
        if flag != 0:
            score = get_pdf_matching_score(
                EPDF, Pre_PDF
            )  # TODO: How to match these two
            print(score)
            score_seq.append(score)
            if score <= thre:
                print("optimal scale is %d:" % (pss - 1))
                return (pss - 1, score_seq)
        Pre_PDF = EPDF
        flag = 1
    return (stopping, score_seq)


def get_max_noise_in_maps(lm, chn=3):
    """
    Description: To find out the maximum level of noise level in the images
    ----------
    [Input]
    a multi-channel tensor of noise map

    [Output]
    A list of  noise level value
    """
    lm_numpy = lm.data.cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (0, 2, 3, 1))
    nl_list = np.zeros((lm_numpy.shape[0], chn, 1))
    for n in range(lm_numpy.shape[0]):
        for c in range(chn):
            nl = np.amax(lm_numpy[n, :, :, c])
            nl_list[n, c] = nl
    return nl_list


def get_smooth_maps(lm, dilk=50, gsd=10):
    """
    Description: To return the refined maps after dilation and gaussian blur
    [Input] a multi-channel tensor of noise map
    [Output] a multi-channel tensor of refined noise map
    """
    kernel = np.ones((dilk, dilk))
    lm_numpy = lm.data.squeeze(0).cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (1, 2, 0))
    ref_lm_numpy = lm_numpy.copy()  # a refined map
    for c in range(lm_numpy.shape[2]):
        nmap = lm_numpy[:, :, c]
        nmap_dilation = cv2.dilate(nmap, kernel, iterations=1)
        ref_lm_numpy[:, :, c] = nmap_dilation
        # ref_lm_numpy[:, :, c] = scipy.ndimage.filters.gaussian_filter(nmap_dilation, gsd)
    RF_tensor = np2ts(ref_lm_numpy)
    RF_tensor = Variable(RF_tensor, volatile=True)


def zeroing_out_maps(lm, keep=0):
    """
    Only Keep one channel and zero out other channels
    [Input] a multi-channel tensor of noise map
    [Output] a multi-channel tensor of noise map after zeroing out items
    """
    lm_numpy = lm.data.squeeze(0).cpu().numpy()
    lm_numpy = np.transpose(lm_numpy, (1, 2, 0))
    ref_lm_numpy = lm_numpy.copy()  # a refined map
    for c in range(lm_numpy.shape[2]):
        if np.isin(c, keep) == 0:
            ref_lm_numpy[:, :, c] = 0.0
    print(ref_lm_numpy)
    RF_tensor = np2ts(ref_lm_numpy)
    RF_tensor = Variable(RF_tensor, volatile=True)
    return RF_tensor


def level_refine(NM_tensor, ref_mode, chn=3):
    """
    Description: To refine the estimated noise level maps
    [Input] the noise map tensor, and a refinement mode
    Mode:
    [0] Get the most salient (the most frequent estimated noise level)
    [1] Get the maximum value of noise level
    [2] Gaussian smooth the noise level map to make the regional estimation more smooth
    [3] Get the average maximum value of the noise level
    [5] Get the CDF thresholded value 
    
    [Output] a refined map tensor with four channels
    """
    # RF_tensor = NM_tensor.clone()  #get a clone version of NM tensor without changing the original one
    if (
        ref_mode == 0 or ref_mode == 1 or ref_mode == 4 or ref_mode == 5
    ):  # if we use a single value for the map
        if ref_mode == 0 or ref_mode == 4:
            nl_list = get_salient_noise_in_maps(NM_tensor, 0.0, chn)

            if ref_mode == 4:  # half the estimation
                nl_list = nl_list - nl_list
            print(nl_list)
        elif ref_mode == 1:
            nl_list = get_max_noise_in_maps(NM_tensor, chn)
        elif ref_mode == 5:
            nl_list = get_cdf_noise_in_maps(NM_tensor, 0.999, chn)

        noise_map = np.zeros(
            (NM_tensor.shape[0], chn, NM_tensor.size()[2], NM_tensor.size()[3])
        )  # initialize the noise map before concatenating
        for n in range(NM_tensor.shape[0]):
            noise_map[n, :, :, :] = np.reshape(
                np.tile(nl_list[n], NM_tensor.size()[2] * NM_tensor.size()[3]),
                (chn, NM_tensor.size()[2], NM_tensor.size()[3]),
            )
        RF_tensor = torch.from_numpy(noise_map).type(torch.FloatTensor)
        RF_tensor = Variable(RF_tensor, volatile=True)

    elif ref_mode == 2:
        RF_tensor = get_smooth_maps(NM_tensor, 10, 5)
    elif ref_mode == 3:
        lb = get_salient_noise_in_maps(NM_tensor)
        up = get_max_noise_in_maps(NM_tensor)
        nl_list = (lb + up) * 0.5
        noise_map = np.zeros(
            (1, chn, NM_tensor.size()[2], NM_tensor.size()[3])
        )  # initialize the noise map before concatenating
        noise_map[0, :, :, :] = np.reshape(
            np.tile(nl_list, NM_tensor.size()[2] * NM_tensor.size()[3]),
            (chn, NM_tensor.size()[2], NM_tensor.size()[3]),
        )
        RF_tensor = torch.from_numpy(noise_map).type(torch.FloatTensor)
        RF_tensor = Variable(RF_tensor, volatile=True)

    return (RF_tensor, nl_list)


def normalize(a, len_v, min_v, max_v):
    """
    normalize the sequence of factors
    """
    norm_a = np.reshape(a, (len_v, 1))
    norm_a = (norm_a - float(min_v)) / float(max_v - min_v)
    return norm_a


def generate_training_noisy_image(current_image, s_or_m, limit_set, c, val=0):
    noise_level_list = np.zeros((c, 1))
    if s_or_m == 0:  # single noise type
        if val == 0:
            for chn in range(c):
                noise_level_list[chn] = np.random.uniform(
                    limit_set[0][0], limit_set[0][1]
                )
        elif val == 1:
            for chn in range(c):
                noise_level_list[chn] = 35
        noisy_img = generate_noisy(current_image, 0, noise_level_list / 255.0)

    return (noisy_img, noise_level_list)


def generate_ground_truth_noise_map(
    noise_map, n, noise_level_list, limit_set, c, pn, pw, ph
):
    for chn in range(c):
        noise_level_list[chn] = normalize(
            noise_level_list[chn], 1, limit_set[0][0], limit_set[0][1]
        )  # normalize the level value
    noise_map[n, :, :, :] = np.reshape(
        np.tile(noise_level_list, pw * ph), (c, pw, ph)
    )  # total number of channels
    return noise_map


# Add noise to the original images
def generate_noisy(image, noise_type, noise_level_list=0, sigma_s=20, sigma_c=40):
    """
    Description: To generate noisy images of different types
    ----------
    [Input]
    image : ndarray of float type: [0,1] just one image, current support gray or color image input (w,h,c)
    noise_type: 0,1,2,3
    noise_level_list: pre-defined noise level for each channel, without normalization: only information of 3 channels
    [0]'AWGN'     Multi-channel Gaussian-distributed additive noise
    [1]'RVIN'    Replaces random pixels with 0 or 1.  noise_level: ratio of the occupation of the changed pixels
    [2]'Gaussian-Poisson'   GP noise approximator, the combinatin of signal-dependent and signal independent noise
    [Output]
    A noisy image
    """
    w, h, c = image.shape
    # Some unused noise type: Poisson and Uniform
    # if noise_type == *:
    # vals = len(np.unique(image))
    # vals = 2 ** np.ceil(np.log2(vals))
    # noisy = np.random.poisson(image * vals) / float(vals)

    # if noise_type == *:
    # uni = np.random.uniform(-factor,factor,(w, h, c))
    # uni = uni.reshape(w, h, c)
    # noisy = image + uni

    noisy = image.copy()

    if noise_type == 0:  # MC-AWGN model
        gauss = np.zeros((w, h, c))
        for chn in range(c):
            gauss[:, :, chn] = np.random.normal(0, noise_level_list[chn], (w, h))
        noisy = image + gauss
    elif noise_type == 1:  # MC-RVIN model
        for chn in range(c):  # process each channel separately
            prob_map = np.random.uniform(0.0, 1.0, (w, h))
            noise_map = np.random.uniform(0.0, 1.0, (w, h))
            noisy_chn = noisy[:, :, chn]
            noisy_chn[prob_map < noise_level_list[chn]] = noise_map[
                prob_map < noise_level_list[chn]
            ]

    elif noise_type == 2:
        pass

    return noisy


# generate AWGN-RVIN noise together
def generate_comp_noisy(image, noise_level_list):

    """
    Description: To generate mixed AWGN and RVIN noise together
    ----------
    [Input]
    image: a float image between [0,1]
    noise_level_list: AWGN and RVIN noise level
    [Output]
    A noisy image
    """
    w, h, c = image.shape
    noisy = image.copy()
    for chn in range(c):
        mix_thre = noise_level_list[c + chn]  # get the mix ratio of AWGN and RVIN
        gau_std = noise_level_list[chn]  # get the gaussian std
        prob_map = np.random.uniform(0, 1, (w, h))  # the prob map
        noise_map = np.random.uniform(0, 1, (w, h))  # the noisy map
        noisy_chn = noisy[:, :, chn]
        noisy_chn[prob_map < mix_thre] = noise_map[prob_map < mix_thre]
        gauss = np.random.normal(0, gau_std, (w, h))
        noisy_chn[prob_map >= mix_thre] = (
            noisy_chn[prob_map >= mix_thre] + gauss[prob_map >= mix_thre]
        )

    return noisy


def generate_denoise(image, model, noise_level_list):
    """
    Description: Generate Denoised Blur Images
    ----------
    [Input]
    image: 
    model: 
    noise_level_list: 
    
    [Output]
    A blur image patch
    """
    # input images
    ISource = np2ts(image)
    ISource = torch.clamp(ISource, 0.0, 1.0)
    ISource = Variable(ISource, volatile=True)
    # input denoise conditions
    noise_map = np.zeros(
        (1, 6, image.shape[0], image.shape[1])
    )  # initialize the noise map before concatenating
    noise_map[0, :, :, :] = np.reshape(
        np.tile(noise_level_list, image.shape[0] * image.shape[1]),
        (6, image.shape[0], image.shape[1]),
    )
    NM_tensor = torch.from_numpy(noise_map).type(torch.FloatTensor)
    NM_tensor = Variable(NM_tensor, volatile=True)
    # generate blur images
    Res = model(ISource, NM_tensor)
    Out = torch.clamp(ISource - Res, 0.0, 1.0)
    out_numpy = Out.data.squeeze(0).cpu().numpy()
    out_numpy = np.transpose(out_numpy, (1, 2, 0))
    return out_numpy


# TODO: two pixel shuffle functions to process the images
def pixelshuffle(image, scale):
    """
    Discription: Given an image, return a reversible sub-sampling
    [Input]: Image ndarray float
    [Return]: A mosic image of shuffled pixels
    """
    if scale == 1:
        return image
    w, h, c = image.shape
    mosaic = np.array([])
    for ws in range(scale):
        band = np.array([])
        for hs in range(scale):
            temp = image[ws::scale, hs::scale, :]  # get the sub-sampled image
            band = np.concatenate((band, temp), axis=1) if band.size else temp
        mosaic = np.concatenate((mosaic, band), axis=0) if mosaic.size else band
    return mosaic


def reverse_pixelshuffle(image, scale, fill=0, fill_image=0, ind=[0, 0]):
    """
    Discription: Given a mosaic image of subsampling, recombine it to a full image
    [Input]: Image
    [Return]: Recombine it using different portions of pixels
    """
    w, h, c = image.shape
    real = np.zeros((w, h, c))  # real image
    wf = 0
    hf = 0
    for ws in range(scale):
        hf = 0
        for hs in range(scale):
            temp = real[ws::scale, hs::scale, :]
            wc, hc, cc = temp.shape  # get the shpae of the current images
            if fill == 1 and ws == ind[0] and hs == ind[1]:
                real[ws::scale, hs::scale, :] = fill_image[
                    wf : wf + wc, hf : hf + hc, :
                ]
            else:
                real[ws::scale, hs::scale, :] = image[wf : wf + wc, hf : hf + hc, :]
            hf = hf + hc
        wf = wf + wc
    return real


def scal2map(level, h, w, min_v=0.0, max_v=255.0):
    """
    Change a single normalized noise level value to a map
    [Input]: level: a scaler noise level(0-1), h, w
    [Return]: a pytorch tensor of the cacatenated noise level map
    """
    # get a tensor from the input level
    level_tensor = torch.from_numpy(np.reshape(level, (1, 1))).type(torch.FloatTensor)
    # make the noise level to a map
    level_tensor = level_tensor.view(stdN_tensor.size(0), stdN_tensor.size(1), 1, 1)
    level_tensor = level_tensor.repeat(1, 1, h, w)
    return level_tensor


def scal2map_spatial(level1, level2, h, w):
    stdN_t1 = scal2map(level1, int(h / 2), w)
    stdN_t2 = scal2map(level2, h - int(h / 2), w)
    stdN_tensor = torch.cat([stdN_t1, stdN_t2], dim=2)
    return stdN_tensor


# the limitation range of each type of noise level: [0]Gaussian [1]Impulse
limit_set = [[0, 75], [0, 80]]


def img_normalize(data):
    return data / 255.0


def denoiser(Img, c, pss, model, model_est, opt):

    w, h, _ = Img.shape
    Img = pixelshuffle(Img, pss)

    noise_level_list = np.zeros((2 * c, 1))  # two noise types with three channels
    if (
        opt.cond == 0
    ):  # if we use the ground truth of noise for denoising, and only one single noise type
        noise_level_list = np.array(opt.test_noise_level)
    elif opt.cond == 2:  # if we use an external fixed input condition for denoising
        noise_level_list = np.array(opt.ext_test_noise_level)

    # Clean Image Tensor for evaluation
    ISource = np2ts(Img)
    # noisy image and true residual
    if (
        opt.real_n == 0 and opt.spat_n == 0
    ):  # no spatial noise setting, and synthetic noise
        noisy_img = generate_comp_noisy(Img, np.array(opt.test_noise_level) / 255.0)
        if opt.color == 0:
            noisy_img = np.expand_dims(noisy_img[:, :, 0], 2)
    elif opt.real_n == 1 or opt.real_n == 2:  # testing real noisy images
        noisy_img = Img
    elif opt.spat_n == 1:
        noisy_img = generate_noisy(Img, 2, 0, 20, 40)
    INoisy = np2ts(noisy_img, opt.color)
    INoisy = torch.clamp(INoisy, 0.0, 1.0)
    True_Res = INoisy - ISource
    ISource, INoisy, True_Res = (
        Variable(ISource, volatile=True),
        Variable(INoisy, volatile=True),
        Variable(True_Res, volatile=True),
    )

    if opt.mode == "MC":
        # obtain the corrresponding input_map
        if (
            opt.cond == 0 or opt.cond == 2
        ):  # if we use ground choose level or the given fixed level
            # normalize noise leve map to [0,1]
            noise_level_list_n = np.zeros((2 * c, 1))
            print(c)
            for noise_type in range(2):
                for chn in range(c):
                    noise_level_list_n[noise_type * c + chn] = normalize(
                        noise_level_list[noise_type * 3 + chn],
                        1,
                        limit_set[noise_type][0],
                        limit_set[noise_type][1],
                    )  # normalize the level value
            # generate noise maps
            noise_map = np.zeros(
                (1, 2 * c, Img.shape[0], Img.shape[1])
            )  # initialize the noise map
            noise_map[0, :, :, :] = np.reshape(
                np.tile(noise_level_list_n, Img.shape[0] * Img.shape[1]),
                (2 * c, Img.shape[0], Img.shape[1]),
            )
            NM_tensor = torch.from_numpy(noise_map).type(torch.FloatTensor)
            NM_tensor = Variable(NM_tensor, volatile=True)
        # use the estimated noise-level map for blind denoising
        elif opt.cond == 1:  # if we use the estimated map directly
            NM_tensor = torch.clamp(model_est(INoisy), 0.0, 1.0)
            if (
                opt.refine == 1
            ):  # if we need to refine the map before putting it to the denoiser
                NM_tensor_bundle = level_refine(
                    NM_tensor, opt.refine_opt, 2 * c
                )  # refine_opt can be max, freq and their average
                NM_tensor = NM_tensor_bundle[0]
                noise_estimation_table = np.reshape(NM_tensor_bundle[1], (2 * c,))
            if opt.zeroout == 1:
                NM_tensor = zeroing_out_maps(NM_tensor, opt.keep_ind)
        Res = model(INoisy, NM_tensor)

    elif opt.mode == "B":
        Res = model(INoisy)

    Out = torch.clamp(INoisy - Res, 0.0, 1.0)  # Output image after denoising

    # get the maximum denoising result
    max_NM_tensor = level_refine(NM_tensor, 1, 2 * c)[0]
    max_Res = model(INoisy, max_NM_tensor)
    max_Out = torch.clamp(INoisy - max_Res, 0.0, 1.0)
    max_out_numpy = visual_va2np(
        max_Out, opt.color, opt.ps, pss, 1, opt.rescale, w, h, c
    )
    del max_Out
    del max_Res
    del max_NM_tensor

    if (opt.ps == 1 or opt.ps == 2) and pss != 1:  # pixelshuffle multi-scale
        # create batch of images with one subsitution
        mosaic_den = visual_va2np(Out, opt.color, 1, pss, 1, opt.rescale, w, h, c)
        out_numpy = np.zeros((pss ** 2, c, w, h))
        # compute all the images in the ps scale set
        for row in range(pss):
            for column in range(pss):
                re_test = (
                    visual_va2np(
                        Out,
                        opt.color,
                        1,
                        pss,
                        1,
                        opt.rescale,
                        w,
                        h,
                        c,
                        1,
                        visual_va2np(INoisy, opt.color),
                        [row, column],
                    )
                    / 255.0
                )
                # cv2.imwrite(os.path.join(opt.out_dir,file_name + '_%d_%d.png' % (row, column)), re_test[:,:,::-1]*255.)
                re_test = np.expand_dims(re_test, 0)
                if opt.color == 0:  # if gray image
                    re_test = np.expand_dims(re_test[:, :, :, 0], 3)
                re_test_tensor = torch.from_numpy(
                    np.transpose(re_test, (0, 3, 1, 2))
                ).type(torch.FloatTensor)
                re_test_tensor = Variable(re_test_tensor, volatile=True)
                re_NM_tensor = torch.clamp(model_est(re_test_tensor), 0.0, 1.0)
                if (
                    opt.refine == 1
                ):  # if we need to refine the map before putting it to the denoiser
                    re_NM_tensor_bundle = level_refine(
                        re_NM_tensor, opt.refine_opt, 2 * c
                    )  # refine_opt can be max, freq and their average
                    re_NM_tensor = re_NM_tensor_bundle[0]
                re_Res = model(re_test_tensor, re_NM_tensor)
                Out2 = torch.clamp(re_test_tensor - re_Res, 0.0, 1.0)
                out_numpy[row * pss + column, :, :, :] = Out2.data.cpu().numpy()
                del Out2
                del re_Res
                del re_test_tensor
                del re_NM_tensor
                del re_test

        out_numpy = np.mean(out_numpy, 0)
        out_numpy = np.transpose(out_numpy, (1, 2, 0)) * 255.0
    elif opt.ps == 0 or pss == 1:  # other cases
        out_numpy = visual_va2np(Out, opt.color, 0, 1, 1, opt.rescale, w, h, c)

    out_numpy = out_numpy.astype(np.float32)  # details
    max_out_numpy = max_out_numpy.astype(np.float32)  # background

    # merging the details and background to balance the effect
    k = opt.k
    merge_out_numpy = (1 - k) * out_numpy + k * max_out_numpy
    merge_out_numpy = merge_out_numpy.astype(np.float32)

    return merge_out_numpy



from typing import Literal, Optional

import torch
from skimage.exposure import match_histograms


def spectral_reducer(
    X: torch.Tensor, method: Literal["mean", "median", "max", "min"] = "mean"
) -> torch.Tensor:
    """ Reduce the number of channels of a tensor from (C, H, W) to 
    (H, W) using a given method.
    
    Args:
        X (torch.Tensor): The tensor to reduce.
        
        method (str, optional): The method used to reduce the number of 
            channels. Must be one of "mean", "median", "max", "min", 
            "luminosity". Defaults to "mean".
            
    Raises:
        ValueError: If the method is not valid.
    
    Returns:
        torch.Tensor: The reduced tensor.
    """
    if method == "mean":
        return X.mean(dim=0)
    elif method == "median":
        return X.median(dim=0).values
    elif method == "max":
        return X.max(dim=0).values
    elif method == "min":
        return X.min(dim=0).values
    else:
        raise ValueError(
            "Invalid method. Must be one of 'mean', 'median', 'max', 'min', 'luminosity'."
        )


def spatial_reducer(
    x: torch.Tensor,
    reduction: Literal[
        "mean_abs", "mean", "median", "median_abs", "max", "max_abs", "min", "min_abs"
    ],
) -> torch.Tensor:
    """Reduces a given tensor by a given reduction method.
    
    Args:
        x (torch.Tensor): The tensor to reduce
        reduction (Literal["mean_abs", "mean", "median", 
            "median_abs", "max","max_abs", "min", "min_abs"]): 
            The reduction method to use.
                
    Return:
        torch.Tensor: The reduced tensor
        
    Raise:
        ValueError: If the reduction method is not supported.
    """
    if reduction == "mean":
        return torch.nanmean(x)
    if reduction == "mean_abs":
        return torch.nanmean(torch.abs(x))
    if reduction == "median":
        return torch.nanmedian(x)
    if reduction == "median_abs":
        return torch.nanmedian(torch.abs(x))
    if reduction == "max":
        return torch.max(x[x == x])
    if reduction == "max_abs":
        xabs = torch.abs(x)
        return torch.max(xabs[xabs == xabs])
    if reduction == "min":
        return torch.min(x[x == x])
    if reduction == "min_abs":
        xabs = torch.abs(x)
        return torch.min(xabs[xabs == xabs])
    if reduction == "sum":
        return torch.nansum(x)
    if reduction == "sum_abs":
        return torch.nansum(torch.abs(x))
    if reduction == "none" or reduction is None:
        raise ValueError("None is not a valid reduction method.")

    raise ValueError("Reduction parameter unknown.")


def hq_histogram_matching(image1: torch.Tensor, image2: torch.Tensor) -> torch.Tensor:
    """ Lazy implementation of histogram matching 

    Args:
        image1 (torch.Tensor): The low-resolution image (C, H, W).
        image2 (torch.Tensor): The super-resolved image (C, H, W).

    Returns:
        torch.Tensor: The super-resolved image with the histogram of
            the target image.
    """

    # Go to numpy
    np_image1 = image1.detach().cpu().numpy()
    np_image2 = image2.detach().cpu().numpy()

    if np_image1.ndim == 3:
        np_image1_hat = match_histograms(np_image1, np_image2, channel_axis=0)
    elif np_image1.ndim == 2:
        np_image1_hat = match_histograms(np_image1, np_image2, channel_axis=None)
    else:
        raise ValueError("The input image must have 2 or 3 dimensions.")

    # Go back to torch
    image1_hat = torch.from_numpy(np_image1_hat).to(image1.device)

    return image1_hat


def estimate_medoid(tensor: torch.Tensor) -> torch.Tensor:
    """ Estimate the medoid of a tensor."""
    # Compute the pairwise distances between the elements of the tensor
    distances = torch.cdist(tensor[:, None], tensor[:, None])

    # Compute the sum of distances for each element
    sums = distances.sum(dim=1)

    # Select the element with the smallest sum of distances
    medoid_index = sums.argmin()
    medoid = tensor[medoid_index]

    return medoid


def do_square(
    tensor: torch.Tensor,
    patch_size: Optional[int] = 32,
    get_pad: bool = False,
    constant_value: float = torch.nan,
) -> torch.Tensor:
    """ Split a tensor into n_patches x n_patches patches and return
    the patches as a tensor.
    
    Args:
        tensor (torch.Tensor): The tensor to split.
        n_patches (int, optional): The number of patches to split the tensor into.
            If None, the tensor is split into the smallest number of patches.

    Returns:
        torch.Tensor: The patches as a tensor.
    """
    # tensor = torch.rand(3, 100, 100)
    # Check if it is a square tensor
    if tensor.shape[-1] != tensor.shape[-2]:
        raise ValueError("The tensor must be square.")

    if patch_size is None:
        patch_size = 1

    # tensor (C, H, W)
    xdim = tensor.shape[1]
    ydim = tensor.shape[2]

    minimages_x = int(torch.ceil(torch.tensor(xdim / patch_size)))
    minimages_y = int(torch.ceil(torch.tensor(ydim / patch_size)))

    pad_x_01 = int((minimages_x * patch_size - xdim) // 2)
    pad_x_02 = int((minimages_x * patch_size - xdim) - pad_x_01)

    pad_y_01 = int((minimages_y * patch_size - ydim) // 2)
    pad_y_02 = int((minimages_y * patch_size - ydim) - pad_y_01)

    padded_tensor = torch.nn.functional.pad(
        input=tensor,
        pad=(pad_x_01, pad_x_02, pad_y_01, pad_y_02),
        mode="constant",
        value=constant_value,
    )

    # split the tensor (C, H, W) into (n_patches, n_patches, C, H, W)
    patches = padded_tensor.unfold(1, patch_size, patch_size).unfold(
        2, patch_size, patch_size
    )

    # move the axes (C, n_patches, n_patches, H, W) -> (n_patches, n_patches, C, H, W)
    patches = patches.permute(1, 2, 0, 3, 4)

    # compress dimensions (n_patches, n_patches, C, H, W) -> (n_patches*n_patches, C, H, W)
    patches = patches.reshape(-1, *patches.shape[2:])

    if get_pad:
        return patches, (pad_x_01, pad_x_02, pad_y_01, pad_y_02)

    return patches


def get_numbers(string: str) -> str:
    """ Get the numbers from a string. """
    n_patches = string.replace("patch", "")
    if n_patches == "":
        n_patches = 32
    else:
        try:
            n_patches = int(n_patches)
        except:
            raise ValueError("Invalid patch name.")
    return n_patches


def check_lpips():
    try:
        import lpips
    except ImportError as e:
        raise ImportError(
            "The extra requirement for perceptual metrics is not installed. "
            "Please install it with `pip install opensr-test[perceptual]`."
        ) from e

def check_openclip():
    try:
        import open_clip
    except ImportError as e:
        raise ImportError(
            "The extra requirement for perceptual metrics is not installed. "
            "Please install it with `pip install opensr-test[perceptual]`."
        ) from e


import collections.abc as collections
from pathlib import Path
from types import SimpleNamespace
from typing import Callable, List, Optional, Tuple, Union

import cv2
import kornia
import numpy as np
import torch


class ImagePreprocessor:
    default_conf = {
        "resize": None,  # target edge length, None for no resizing
        "side": "long",
        "interpolation": "bilinear",
        "align_corners": None,
        "antialias": True,
        "grayscale": False,  # convert rgb to grayscale
    }

    def __init__(self, **conf) -> None:
        super().__init__()
        self.conf = {**self.default_conf, **conf}
        self.conf = SimpleNamespace(**self.conf)

    def __call__(self, img: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Resize and preprocess an image, return image and resize scale"""
        h, w = img.shape[-2:]
        if self.conf.resize is not None:
            img = kornia.geometry.transform.resize(
                img,
                self.conf.resize,
                side=self.conf.side,
                antialias=self.conf.antialias,
                align_corners=self.conf.align_corners,
            )
        scale = torch.Tensor([img.shape[-1] / w, img.shape[-2] / h]).to(img)
        if self.conf.grayscale and img.shape[-3] == 3:
            img = kornia.color.rgb_to_grayscale(img)
        elif not self.conf.grayscale and img.shape[-3] == 1:
            img = kornia.color.grayscale_to_rgb(img)
        return img, scale


def map_tensor(input_, func: Callable):
    string_classes = (str, bytes)
    if isinstance(input_, string_classes):
        return input_
    elif isinstance(input_, collections.Mapping):
        return {k: map_tensor(sample, func) for k, sample in input_.items()}
    elif isinstance(input_, collections.Sequence):
        return [map_tensor(sample, func) for sample in input_]
    elif isinstance(input_, torch.Tensor):
        return func(input_)
    else:
        return input_


def batch_to_device(batch: dict, device: str = "cpu", non_blocking: bool = True):
    """Move batch (dict) to device"""

    def _func(tensor):
        return tensor.to(device=device, non_blocking=non_blocking).detach()

    return map_tensor(batch, _func)


def rbd(data: dict) -> dict:
    """Remove batch dimension from elements in data"""
    return {
        k: v[0] if isinstance(v, (torch.Tensor, np.ndarray, list)) else v
        for k, v in data.items()
    }


def read_image(path: Path, grayscale: bool = False) -> np.ndarray:
    """Read an image from path as RGB or grayscale"""
    if not Path(path).exists():
        raise FileNotFoundError(f"No image at path {path}.")
    mode = cv2.IMREAD_GRAYSCALE if grayscale else cv2.IMREAD_COLOR
    image = cv2.imread(str(path), mode)
    if image is None:
        raise IOError(f"Could not read image at {path}.")
    if not grayscale:
        image = image[..., ::-1]
    return image


def numpy_image_to_torch(image: np.ndarray) -> torch.Tensor:
    """Normalize the image tensor and reorder the dimensions."""
    if image.ndim == 3:
        image = image.transpose((2, 0, 1))  # HxWxC to CxHxW
    elif image.ndim == 2:
        image = image[None]  # add channel axis
    else:
        raise ValueError(f"Not an image: {image.shape}")
    return torch.tensor(image / 255.0, dtype=torch.float)


def resize_image(
    image: np.ndarray,
    size: Union[List[int], int],
    fn: str = "max",
    interp: Optional[str] = "area",
) -> np.ndarray:
    """Resize an image to a fixed size, or according to max or min edge."""
    h, w = image.shape[:2]

    fn = {"max": max, "min": min}[fn]
    if isinstance(size, int):
        scale = size / fn(h, w)
        h_new, w_new = int(round(h * scale)), int(round(w * scale))
        scale = (w_new / w, h_new / h)
    elif isinstance(size, (tuple, list)):
        h_new, w_new = size
        scale = (w_new / w, h_new / h)
    else:
        raise ValueError(f"Incorrect new size: {size}")
    mode = {
        "linear": cv2.INTER_LINEAR,
        "cubic": cv2.INTER_CUBIC,
        "nearest": cv2.INTER_NEAREST,
        "area": cv2.INTER_AREA,
    }[interp]
    return cv2.resize(image, (w_new, h_new), interpolation=mode), scale


def load_image(path: Path, resize: int = None, **kwargs) -> torch.Tensor:
    image = read_image(path)
    if resize is not None:
        image, _ = resize_image(image, resize, **kwargs)
    return numpy_image_to_torch(image)


def match_pair(
    extractor,
    matcher,
    image0: torch.Tensor,
    image1: torch.Tensor,
    device: str = "cpu",
    **preprocess,
):
    """Match a pair of images (image0, image1) with an extractor and matcher"""
    feats0 = extractor.extract(image0, **preprocess)
    feats1 = extractor.extract(image1, **preprocess)
    matches01 = matcher({"image0": feats0, "image1": feats1})
    data = [feats0, feats1, matches01]
    # remove batch dim and move to target device
    feats0, feats1, matches01 = [batch_to_device(rbd(x), device) for x in data]
    return feats0, feats1, matches01



from argparse import Namespace

import pkg_resources
import rasterio as rio
import torch
from opensr_test.denoiser.utils import (NoisyEstimator, RealNoisyEstimator,
                                        denoiser)


def setup_denoiser_model(device=torch.device("cpu")):

    # get the package directory
    weights01 = pkg_resources.resource_filename(
        "opensr_test", "denoiser/weights/NoisyEstimator.pth"
    )
    weights02 = pkg_resources.resource_filename(
        "opensr_test", "denoiser/weights/RealNoisyEstimator.pth"
    )

    # Load Noisy Calibration Model
    model = NoisyEstimator(channels=3, num_of_layers=20, num_of_est=6)
    weights = torch.load(weights01)
    model.load_state_dict(weights)
    model.eval()
    model.to(device)

    model_real = RealNoisyEstimator(input_channels=3, output_channels=6)
    weights = {k.replace("module.", ""): v for k, v in torch.load(weights02).items()}
    model_real.load_state_dict(weights)
    model_real.eval()
    model_real.to(device)

    return {"noisy_estimator": model, "real_noisy_estimator": model_real}


def remove_noise(img, models):
    # all the parameters
    parameters = {
        "color": 1,
        "cond": 1,
        "ext_test_noise_level": None,
        "k": 0,
        "keep_ind": [0, 1, 2],
        "mode": "MC",
        "num_of_layers": 20,
        "output_map": 0,
        "ps": 2,
        "ps_scale": 2,
        "real_n": 1,
        "refine": 0,
        "refine_opt": 1,
        "rescale": 1,
        "scale": 1,
        "spat_n": 0,
        "test_noise_level": None,
        "wbin": 64,
        "zeroout": 0,
    }

    # from dict to namespace
    opt = Namespace(**parameters)
    pss = opt.ps_scale
    model, model_est = models.values()

    return denoiser(img, 3, pss, model, model_est, opt) / 255



import warnings
from typing import Any, Optional, Union

import opensr_test.plot
import torch
from opensr_test.config import (
    Auxiliar, Consistency, Distance,
    Results, Correctness, Config
)
from opensr_test.hallucinations import get_distances, tc_improvement, tc_omission, tc_hallucination
from opensr_test.kernels import apply_downsampling, apply_upsampling
from opensr_test.reflectance import reflectance_metric
from opensr_test.spatial import (SpatialMetric, spatial_aligment,
                                 spatial_setup_model)
from opensr_test.spectral import spectral_metric
from opensr_test.utils import hq_histogram_matching


class Metrics:
    def __init__(
        self,
        device: Union[str, torch.device, None] = "cpu",
        **kwargs: Any
    ) -> None:
        """ A class to evaluate the performance of a image 
        enhancement algorithm considering the triplets: LR[input], 
        SR[enhanced], HR[ground truth].

        Args:
            params (Optional[Config], optional): The parameters to
                setup the opensr-test experiment. Defaults to None.
                If None, the default parameters are used. See 
                config.py for more information.
            device (Union[str, torch.device, None], optional): The
                device to use. Defaults to "cpu".
        """
        

        # Set the parameters
        if kwargs is None:
            self.params = Config()
        else:
            self.params = Config(**kwargs)
        
        # If patch size is 1, then the aggregation method must be pixel
        if self.params.patch_size == 1:
            self.params.agg_method = "pixel"

        # Global parameters
        self.method = self.params.agg_method
        self.device = device

        

        # Set the spatial grid regulator
        self.apply_upsampling = apply_upsampling
        self.apply_downsampling = apply_downsampling

        # Setup the spatial model
        self.spatial_model = spatial_setup_model(
            features=self.params.spatial_features,
            matcher=self.params.spatial_matcher,
            max_num_keypoints=self.params.spatial_max_num_keypoints,
            device=device,
        )

        # Initial triplets: LR[input], SR[enhanced], HR[ground truth]
        self.lr = None
        self.sr = None
        self.hr = None
        
        # The sr without systematic error (spectral & spatial removed)
        self.sr_harm = None

        # The LR in the HR space: (C, H, W) -> (H*scale, W*scale)
        # using a bilinear interpolation (zero-parameter model)
        # also applied the spectral_reducer since we don't need
        # to evaluate the spectral information in the HR space.
        self.lr_to_hr = None

        # The SR in the LR space: (C, H*scale, W*scale) -> (C, H, W)
        # using a bilinear interpolation (zero-parameter model).
        self.sr_to_lr = None

        # The landuse map (H, W)
        self.landuse = None

        # The scale value of the experiment. Obtained from the
        # ratio between the HR and LR image sizes.
        self.scale_factor = None

        # RGB pairs (used for plotting and some metrics [spatial
        # & perceptual])
        self.lr_RGB = None
        self.hr_RGB = None
        self.sr_RGB = None

        self.lr_to_hr_RGB = None
        self.sr_to_lr_RGB = None
        self.sr_harm_RGB = None

        # Metrics to be computed

        ## consistency with low resolution
        self.reflectance_value = None
        self.spectral_value = None
        self.spatial_aligment_value = None
        
        ## distance to the omission or improvement space
        self.d_ref = None
        self.d_om = None
        self.d_hr = None
        
        ## Get the energy matrix
        self.energy = None

        ## Correctness metrics
        self.omission = None
        self.improvement = None
        self.hallucination = None
        self.classification = None

        ## Percentage of hallucination and improvement
        self.ha_percentage = None
        self.om_percentage = None
        self.im_percentage = None

    def setup(
        self,
        lr: torch.Tensor,
        sr: torch.Tensor,
        hr: torch.Tensor,
        landuse: Optional[torch.Tensor] = None,
        downsample_method: Optional[str] = "classic",
        upsample_method: Optional[str] = "classic",
    ) -> None:
        """ Obtain the performance metrics of the SR image.

        Args:
            lr (torch.Tensor): The LR image as a tensor (C, H, W).
            sr (torch.Tensor): The SR image as a tensor (C, H, W).
            hr (torch.Tensor): The HR image as a tensor (C, H, W).
        """
        # Check if SR has gradients
        if sr.requires_grad:
            raise ValueError("The SR image must not have gradients.")

        # If patch size is higher than the image size, then
        # return an error.
        if self.method == "patch":
            condition1 = self.params.patch_size > lr.shape[1]
            condition2 = self.params.patch_size > lr.shape[2]
            if condition1 or condition2:
                raise ValueError(
                    "The patch size must be lower than the image size."
                )

        # Obtain the scale factor
        scale_factor = hr.shape[-1] / lr.shape[-1]
        if not scale_factor.is_integer():
            raise ValueError("The scale factor must be an integer.")
        self.scale_factor = int(scale_factor)


        # Move all the images to the same device
        self.lr = self.apply_mask(lr.to(self.device), self.params.mask // self.scale_factor)
        self.sr = self.apply_mask(sr.to(self.device), self.params.mask)
        self.hr = self.apply_mask(hr.to(self.device), self.params.mask)
        self.landuse = self.apply_mask(landuse.to(self.device), self.params.mask // self.scale_factor) if landuse is not None else None


        # Obtain the LR in the HR space
        if self.scale_factor > 1:
            self.lr_to_hr = self.apply_downsampling(
                X=self.lr[None], 
                scale=self.scale_factor, 
                method=downsample_method
            ).squeeze(0)
        else:
            self.lr_to_hr = self.lr

        # Obtain the SR in the LR space
        self.sr_to_lr = self.apply_upsampling(
            X=self.sr[None],
            scale=self.scale_factor,
            method=upsample_method
        ).squeeze(0)

        # Obtain the RGB images
        if self.lr.shape[0] >= 3:
            self.lr_RGB = self.lr[self.params.rgb_bands]
            self.hr_RGB = self.hr[self.params.rgb_bands]
            self.sr_RGB = self.sr[self.params.rgb_bands]
            self.lr_to_hr_RGB = self.lr_to_hr[self.params.rgb_bands]
            self.sr_to_lr_RGB = self.sr_to_lr[self.params.rgb_bands]
        else:
            self.lr_RGB = self.lr[0][None]
            self.hr_RGB = self.hr[0][None]
            self.sr_RGB = self.sr[0][None]
            self.lr_to_hr_RGB = self.lr_to_hr[0][None]
            self.sr_to_lr_RGB = self.sr_to_lr[0][None]

        # Apply harmozination
        self.sr_harm_setup()

        return None

    def sr_harm_setup(self) -> None:
        """ Remove the systematic error from the SR image.
                        
        Returns:
            torch.Tensor: The SR image without systematic error.        
        """

        # Remove systematic reflectance error
        if self.params.harm_apply_spectral:
            sr_harm = hq_histogram_matching(self.sr, self.hr)
        else:
            sr_harm = self.sr

        if self.params.harm_apply_spatial:
            sr_harm, matching_points = spatial_aligment(
                sr=sr_harm,
                hr=self.hr,
                spatial_model=self.spatial_model,
                threshold_n_points=self.params.spatial_threshold_npoints,
                threshold_distance=self.params.spatial_threshold_distance,
                rgb_bands=self.params.rgb_bands,
            )
        else:
            sr_harm = sr_harm

        # Save the SR harmonized image
        self.sr_harm = sr_harm
        self.matching_points_02 = matching_points

        # Estimate mask and apply to HR and SR
        mask = torch.sum(torch.abs(self.sr_harm), axis=0) != 0
        self.hr = self.hr * mask
        self.hr_RGB = self.hr_RGB * mask

        if self.lr.shape[0] >= 3:
            self.sr_harm_RGB = self.sr_harm[self.params.rgb_bands]
        else:
            self.sr_harm_RGB = self.sr_harm[0][None]

    def _reflectance_metric(self) -> None:
        """ Estimate the spectral global error by comparing the
        reflectance of the LR and SR images.
        
        Returns:
            Value: The spectral global error.
        """
        self.reflectance_value = reflectance_metric(
            lr=self.lr,
            sr_to_lr=self.sr_to_lr,
            metric=self.params.reflectance_method,
            agg_method=self.params.agg_method,
            patch_size=self.params.patch_size,
        )

    def _spectral_metric(self) -> None:
        """ Estimate the spectral local error by comparing the
        reflectance of the LR and SR images.

        Returns:
            float: The spectral local error.
        """
        self.spectral_value = spectral_metric(
            lr=self.lr,
            sr_to_lr=self.sr_to_lr,
            metric=self.params.spectral_method,
            agg_method=self.params.agg_method,
            patch_size=self.params.patch_size,
        )

    def _spatial_metric(self) -> None:
        """ Estimate the spatial error by comparing the
        spatial information of the LR and SR images.
        
        Returns:
            Value: The spatial error.
        """
        # Spatial parameters that control the set of
        # point used to fit the LINEAR spatial model.
        threshold_distance = self.params.spatial_threshold_distance
        threshold_npoints = self.params.spatial_threshold_npoints
        spatial_description = "%s & %s" % (
            self.params.spatial_features,
            self.params.spatial_matcher,
        )

        spatial_metric_result = SpatialMetric(
            x=self.lr_RGB,
            y=self.sr_to_lr_RGB,
            spatial_model=self.spatial_model,
            name=spatial_description,
            threshold_n_points=threshold_npoints,
            threshold_distance=threshold_distance,
            method="pixel",
            patch_size=self.params.patch_size,
            device=self.device,
        )

        self.spatial_aligment_value = spatial_metric_result.compute()
        self.matching_points_01 = spatial_metric_result.matching_points

    def _create_mask(self, stability_threshold: float = 0.) -> None:
        d_ref, d_im, d_om = get_distances(
            lr_to_hr=self.lr_to_hr,
            sr_harm=self.sr_harm,
            hr=self.hr,
            distance_method="l1",
            agg_method="pixel"
        )

        # create mask
        mask1 = (d_ref > stability_threshold)*1
        mask2 = (d_im > stability_threshold)*1
        mask3 = (d_om > stability_threshold)*1
        mask = ((mask1 + mask2 + mask3) > 0) * 1.0
        mask[mask == 0] = torch.nan
        self.tc_mask = mask
        return mask

    def _distance_metric(self, stability_threshold: float = 0.) -> None:
        self.d_ref, self.d_im, self.d_om = get_distances(
            lr_to_hr=self.lr_to_hr,
            sr_harm=self.sr_harm,
            hr=self.hr,
            distance_method=self.params.distance_method,
            agg_method=self.params.agg_method,
            patch_size=self.params.patch_size,
            rgb_bands=self.params.rgb_bands,
            device=self.device,
        )

        # Apply mask
        mask = self._create_mask(stability_threshold)
        self.potential_pixels = torch.nansum(mask)
        self.d_ref_masked = self.d_ref * mask
        self.d_im_masked = self.d_im * mask
        self.d_om_masked = self.d_om * mask

        # Compute relative distance
        self.d_im_ref = self.d_im_masked / self.d_ref_masked
        self.d_om_ref = self.d_om_masked / self.d_ref_masked

        if sum(self.d_im_ref.shape) == 0:
            self.d_im_ref = self.d_im_ref.reshape(-1)
            self.d_om_ref = self.d_om_ref.reshape(-1)

    def _improvement(self, im_score: float = 0.80) -> None:
        self.improvement = tc_improvement(
            d_im=self.d_im_ref,
            d_om=self.d_om_ref,
            plambda=im_score
        )

    def _omission(self, om_score: float = 0.80) -> None:
        self.omission = tc_omission(
            d_im=self.d_im_ref,
            d_om=self.d_om_ref,
            plambda=om_score
        )

    def _hallucination(self, ha_score: float = 0.80) -> None:
        self.hallucination = tc_hallucination(
            d_im=self.d_im_ref,
            d_om=self.d_om_ref,
            plambda=ha_score
        )
        
        correctness_stack = torch.stack([
            self.improvement,
            self.omission,
            self.hallucination
        ], dim=0)
        
        self.classification = torch.argmin(correctness_stack, dim=0) * self.tc_mask
        
        self.im_percentage = torch.sum(self.classification==0) / self.potential_pixels
        self.om_percentage = torch.sum(self.classification==1) / self.potential_pixels
        self.ha_percentage = torch.sum(self.classification==2) / self.potential_pixels

    def _energy(self) -> None:
        self.energy = self.d_om * self.tc_mask

    def _prepare(self) -> None:
        self.results = Results(
            consistency=Consistency(
                reflectance=dict(self.reflectance_value),
                spectral=dict(self.spectral_value),
                spatial=dict(self.spatial_aligment_value),
            ),
            distance=Distance(
                lr_to_hr=self.d_ref, sr_to_hr=self.d_im, sr_to_lr=self.d_om
            ),
            synthesis=self.energy,
            correctness=Correctness(
                omission=self.omission,
                improvement=self.improvement,
                hallucination=self.hallucination,
                classification=self.classification
            ),
            auxiliar=Auxiliar(
                sr_harm=self.sr_harm,
                lr_to_hr=self.lr_to_hr,
                matching_points_lr=self.matching_points_01,
                matching_points_hr=self.matching_points_02,
            ),
        )

    def summary(self) -> dict:
        return {
            "reflectance": float(self.reflectance_value.value.nanmean()),
            "spectral": float(self.spectral_value.value.nanmean()),
            "spatial": float(self.spatial_aligment_value.value.nanmean()),
            "energy": float(self.energy.nanmean()),
            "ha_percent": float(self.ha_percentage),
            "om_percent": float(self.om_percentage),
            "im_percent": float(self.im_percentage),
        }

    def compute(
        self,
        stability_threshold: Optional[float] = 0.01,
        im_score: Optional[float] = 0.8,
        om_score: Optional[float] = 0.8,
        ha_score: Optional[float] = 0.4,
    ) -> None:

        """ Obtain the performance metrics of the SR image.
        
        Args:
            lr (torch.Tensor): The LR image as a tensor (C, H, W).
            sr (torch.Tensor): The SR image as a tensor (C, H, W).
            hr (torch.Tensor): The HR image as a tensor (C, H, W).
            grid (Optional[bool], optional): Whether to return the
                metrics as grids or not. Defaults to False.
            only_value (Optional[bool], optional): Whether to return
                only the values of the metrics or the full Value 
                object (with description). Defaults to True.
                
        Returns:
            dict: The performance metrics for the SR image.
        """
        # Obtain the RS metrics
        self._reflectance_metric()
        self._spectral_metric()
        self._spatial_metric()

        # Create SR' without systematic error
        self.sr_harm_setup()

        # Obtain the distance metrics
        self._distance_metric(stability_threshold)

        # Obtain the energy
        self._energy()

        # Obtain the improvement metrics
        self._improvement(im_score)

        # Obtain the omission metrics
        self._omission(om_score)

        # Obtain the hallucination metrics
        self._hallucination(ha_score)
        
        # Prepare the results
        self._prepare()


        return None
    def plot_triplets(self, apply_harm: bool = True, stretch: Optional[str] = "linear"):
        if apply_harm:
            tplot = opensr_test.plot.triplets(
                lr_img=self.lr_RGB,
                sr_img=self.sr_harm_RGB,
                hr_img=self.hr_RGB,
                stretch=stretch,
            )
        else:
            tplot = opensr_test.plot.triplets(
                lr_img=self.lr_RGB,
                sr_img=self.sr_RGB,
                hr_img=self.hr_RGB,
                stretch=stretch,
            )
        return tplot

    def plot_quadruplets(
        self, apply_harm: bool = True, stretch: Optional[str] = "linear"
    ):
        if apply_harm:
            qplot = opensr_test.plot.quadruplets(
                lr_img=self.lr_RGB,
                sr_img=self.sr_harm_RGB,
                hr_img=self.hr_RGB,
                landuse_img=self.landuse,
                stretch=stretch,
            )
        else:
            qplot = opensr_test.plot.quadruplets(
                lr_img=self.lr_RGB,
                sr_img=self.sr_RGB,
                hr_img=self.hr_RGB,
                landuse_img=self.landuse,
                stretch=stretch,
            )
        return qplot

    def plot_spatial_matches(self, stretch: Optional[str] = "linear"):
        
        # Retrieve the linear affine model and the matching points
        if self.results.auxiliar.matching_points_lr is False:
            warnings.warn("Spatial model is nan. No spatial matches will be plotted.")
            return None

        matching_points = self.results.auxiliar.matching_points_lr

        # plot it!
        return opensr_test.plot.spatial_matches(
            lr=self.lr_RGB,
            sr_to_lr=self.sr_to_lr_RGB,
            points0=matching_points["points0"],
            points1=matching_points["points1"],
            matches01=matching_points["matches01"],
            threshold_distance=self.params.spatial_threshold_distance,
            stretch=stretch,
        )        

    def plot_summary(self, stretch: Optional[str] = "linear"):
        contion1 = self.params.agg_method == "pixel"
        contion2 = self.method == "patch"
        if (contion1 and contion2):
            raise ValueError(
                "This method only works for pixel and patch evaluation."
            )

        # Reflectance metric
        e1 = self.results.consistency.reflectance.value
        e1_title = self.results.consistency.reflectance.description
        e1_subtitle = "%.04f" % float(e1.nanmean())

        # Spectral metric
        e2 = self.results.consistency.spectral.value
        e2_title = self.results.consistency.spectral.description
        e2_subtitle = "%.04f" % float(e2.nanmean())

        # Spatial error
        e3 = self.results.consistency.spatial.value
        e3_p_np = (
            self.results.auxiliar.matching_points_lr["points0"]
            .clone()
            .detach()
            .cpu()
            .numpy()
        )

        # if patch, then divide by the scale factor
        e3_points = [list(x.flatten().astype(int)) for x in e3_p_np]
        e3_title = self.results.consistency.spatial.description
        e3_subtitle = "%.04f" % float(e3.nanmean())

        # High frequency
        e4 = self.results.distance.sr_to_lr
        e4_title = self.params.distance_method.upper()
        e4_subtitle = "%.04f" % float(e4.nanmean())

        # Error grids
        e5 = self.results.distance.sr_to_hr
        e5_title = self.params.distance_method.upper()
        e5_subtitle = "%.04f" % float(e5.nanmean())

        # Plot high frequency
        fig, axs = opensr_test.plot.display_results(
            lr=self.lr_RGB,
            lrdown=self.lr_to_hr_RGB,
            sr=self.sr_RGB,
            srharm=self.sr_harm_RGB,
            hr=self.hr_RGB,
            e1=e1,
            e2=e2,
            e3=e3,
            e4=e4,
            e5=e5,
            e1_title=e1_title,
            e2_title=e2_title,
            e3_title=e3_title,
            e4_title=e4_title,
            e5_title=e5_title,
            e1_subtitle=e1_subtitle,
            e2_subtitle=e2_subtitle,
            e3_subtitle=e3_subtitle,
            e4_subtitle=e4_subtitle,
            e5_subtitle=e5_subtitle,
            e3_points=e3_points,
            stretch=stretch,
        )
        return fig, axs

    def plot_tc(
        self, 
        log_scale: Optional[bool] = True,
        stretch: Optional[str] = "linear"
    ):
        return opensr_test.plot.display_tc_score(
            sr_rgb=self.sr_harm_RGB,
            d_im_ref=self.d_im_ref,
            d_om_ref=self.d_om_ref,
            tc_score=self.classification,
            log_scale=log_scale,
            stretch=stretch
        )

    def __call__(self) -> Any:
        return self.compute()

    def apply_mask(self, X: torch.Tensor, mask: int) -> torch.Tensor:
        if mask is None:
            return X        
        # C, H, W
        if len(X.shape) == 3:
            return X[:, mask: -mask, mask: -mask]
        elif len(X.shape) == 2:
            return X[mask: -mask, mask: -mask]
        else:
            raise ValueError("The tensor must be 2D or 3D.")



import warnings
from pathlib import Path
from types import SimpleNamespace
from typing import Callable, List, Optional

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

try:
    from flash_attn.modules.mha import FlashCrossAttention
except ModuleNotFoundError:
    FlashCrossAttention = None

if FlashCrossAttention or hasattr(F, "scaled_dot_product_attention"):
    FLASH_AVAILABLE = True
else:
    FLASH_AVAILABLE = False

torch.backends.cudnn.deterministic = True


@torch.cuda.amp.custom_fwd(cast_inputs=torch.float32)
def normalize_keypoints(kpts: torch.Tensor, size: torch.Tensor) -> torch.Tensor:
    if isinstance(size, torch.Size):
        size = torch.tensor(size)[None]
    shift = size.float().to(kpts) / 2
    scale = size.max(1).values.float().to(kpts) / 2
    kpts = (kpts - shift[:, None]) / scale[:, None, None]
    return kpts


def rotate_half(x: torch.Tensor) -> torch.Tensor:
    x = x.unflatten(-1, (-1, 2))
    x1, x2 = x.unbind(dim=-1)
    return torch.stack((-x2, x1), dim=-1).flatten(start_dim=-2)


def apply_cached_rotary_emb(freqs: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
    return (t * freqs[0]) + (rotate_half(t) * freqs[1])


class LearnableFourierPositionalEncoding(nn.Module):
    def __init__(self, M: int, dim: int, F_dim: int = None, gamma: float = 1.0) -> None:
        super().__init__()
        F_dim = F_dim if F_dim is not None else dim
        self.gamma = gamma
        self.Wr = nn.Linear(M, F_dim // 2, bias=False)
        nn.init.normal_(self.Wr.weight.data, mean=0, std=self.gamma ** -2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ encode position vector """
        projected = self.Wr(x)
        cosines, sines = torch.cos(projected), torch.sin(projected)
        emb = torch.stack([cosines, sines], 0).unsqueeze(-3)
        return emb.repeat_interleave(2, dim=-1)


class TokenConfidence(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.token = nn.Sequential(nn.Linear(dim, 1), nn.Sigmoid())

    def forward(self, desc0: torch.Tensor, desc1: torch.Tensor):
        """ get confidence tokens """
        return (
            self.token(desc0.detach().float()).squeeze(-1),
            self.token(desc1.detach().float()).squeeze(-1),
        )


class Attention(nn.Module):
    def __init__(self, allow_flash: bool) -> None:
        super().__init__()
        if allow_flash and not FLASH_AVAILABLE:
            warnings.warn(
                "FlashAttention is not available. For optimal speed, "
                "consider installing torch >= 2.0 or flash-attn.",
                stacklevel=2,
            )
        self.enable_flash = allow_flash and FLASH_AVAILABLE
        if allow_flash and FlashCrossAttention:
            self.flash_ = FlashCrossAttention()

    def forward(self, q, k, v) -> torch.Tensor:
        if self.enable_flash and q.device.type == "cuda":
            if FlashCrossAttention:
                q, k, v = [x.transpose(-2, -3) for x in [q, k, v]]
                m = self.flash_(q.half(), torch.stack([k, v], 2).half())
                return m.transpose(-2, -3).to(q.dtype)
            else:  # use torch 2.0 scaled_dot_product_attention with flash
                args = [x.half().contiguous() for x in [q, k, v]]
                with torch.backends.cuda.sdp_kernel(enable_flash=True):
                    return F.scaled_dot_product_attention(*args).to(q.dtype)
        elif hasattr(F, "scaled_dot_product_attention"):
            args = [x.contiguous() for x in [q, k, v]]
            return F.scaled_dot_product_attention(*args).to(q.dtype)
        else:
            s = q.shape[-1] ** -0.5
            attn = F.softmax(torch.einsum("...id,...jd->...ij", q, k) * s, -1)
            return torch.einsum("...ij,...jd->...id", attn, v)


class Transformer(nn.Module):
    def __init__(
        self, embed_dim: int, num_heads: int, flash: bool = False, bias: bool = True
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        assert self.embed_dim % num_heads == 0
        self.head_dim = self.embed_dim // num_heads
        self.Wqkv = nn.Linear(embed_dim, 3 * embed_dim, bias=bias)
        self.inner_attn = Attention(flash)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.ffn = nn.Sequential(
            nn.Linear(2 * embed_dim, 2 * embed_dim),
            nn.LayerNorm(2 * embed_dim, elementwise_affine=True),
            nn.GELU(),
            nn.Linear(2 * embed_dim, embed_dim),
        )

    def _forward(self, x: torch.Tensor, encoding: Optional[torch.Tensor] = None):
        qkv = self.Wqkv(x)
        qkv = qkv.unflatten(-1, (self.num_heads, -1, 3)).transpose(1, 2)
        q, k, v = qkv[..., 0], qkv[..., 1], qkv[..., 2]
        if encoding is not None:
            q = apply_cached_rotary_emb(encoding, q)
            k = apply_cached_rotary_emb(encoding, k)
        context = self.inner_attn(q, k, v)
        message = self.out_proj(context.transpose(1, 2).flatten(start_dim=-2))
        return x + self.ffn(torch.cat([x, message], -1))

    def forward(self, x0, x1, encoding0=None, encoding1=None):
        return self._forward(x0, encoding0), self._forward(x1, encoding1)


class CrossTransformer(nn.Module):
    def __init__(
        self, embed_dim: int, num_heads: int, flash: bool = False, bias: bool = True
    ) -> None:
        super().__init__()
        self.heads = num_heads
        dim_head = embed_dim // num_heads
        self.scale = dim_head ** -0.5
        inner_dim = dim_head * num_heads
        self.to_qk = nn.Linear(embed_dim, inner_dim, bias=bias)
        self.to_v = nn.Linear(embed_dim, inner_dim, bias=bias)
        self.to_out = nn.Linear(inner_dim, embed_dim, bias=bias)
        self.ffn = nn.Sequential(
            nn.Linear(2 * embed_dim, 2 * embed_dim),
            nn.LayerNorm(2 * embed_dim, elementwise_affine=True),
            nn.GELU(),
            nn.Linear(2 * embed_dim, embed_dim),
        )

        if flash and FLASH_AVAILABLE:
            self.flash = Attention(True)
        else:
            self.flash = None

    def map_(self, func: Callable, x0: torch.Tensor, x1: torch.Tensor):
        return func(x0), func(x1)

    def forward(self, x0: torch.Tensor, x1: torch.Tensor) -> List[torch.Tensor]:
        qk0, qk1 = self.map_(self.to_qk, x0, x1)
        v0, v1 = self.map_(self.to_v, x0, x1)
        qk0, qk1, v0, v1 = map(
            lambda t: t.unflatten(-1, (self.heads, -1)).transpose(1, 2),
            (qk0, qk1, v0, v1),
        )
        if self.flash is not None:
            m0 = self.flash(qk0, qk1, v1)
            m1 = self.flash(qk1, qk0, v0)
        else:
            qk0, qk1 = qk0 * self.scale ** 0.5, qk1 * self.scale ** 0.5
            sim = torch.einsum("b h i d, b h j d -> b h i j", qk0, qk1)
            attn01 = F.softmax(sim, dim=-1)
            attn10 = F.softmax(sim.transpose(-2, -1).contiguous(), dim=-1)
            m0 = torch.einsum("bhij, bhjd -> bhid", attn01, v1)
            m1 = torch.einsum("bhji, bhjd -> bhid", attn10.transpose(-2, -1), v0)
        m0, m1 = self.map_(lambda t: t.transpose(1, 2).flatten(start_dim=-2), m0, m1)
        m0, m1 = self.map_(self.to_out, m0, m1)
        x0 = x0 + self.ffn(torch.cat([x0, m0], -1))
        x1 = x1 + self.ffn(torch.cat([x1, m1], -1))
        return x0, x1


def sigmoid_log_double_softmax(
    sim: torch.Tensor, z0: torch.Tensor, z1: torch.Tensor
) -> torch.Tensor:
    """ create the log assignment matrix from logits and similarity"""
    b, m, n = sim.shape
    certainties = F.logsigmoid(z0) + F.logsigmoid(z1).transpose(1, 2)
    scores0 = F.log_softmax(sim, 2)
    scores1 = F.log_softmax(sim.transpose(-1, -2).contiguous(), 2).transpose(-1, -2)
    scores = sim.new_full((b, m + 1, n + 1), 0)
    scores[:, :m, :n] = scores0 + scores1 + certainties
    scores[:, :-1, -1] = F.logsigmoid(-z0.squeeze(-1))
    scores[:, -1, :-1] = F.logsigmoid(-z1.squeeze(-1))
    return scores


class MatchAssignment(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim
        self.matchability = nn.Linear(dim, 1, bias=True)
        self.final_proj = nn.Linear(dim, dim, bias=True)

    def forward(self, desc0: torch.Tensor, desc1: torch.Tensor):
        """ build assignment matrix from descriptors """
        mdesc0, mdesc1 = self.final_proj(desc0), self.final_proj(desc1)
        _, _, d = mdesc0.shape
        mdesc0, mdesc1 = mdesc0 / d ** 0.25, mdesc1 / d ** 0.25
        sim = torch.einsum("bmd,bnd->bmn", mdesc0, mdesc1)
        z0 = self.matchability(desc0)
        z1 = self.matchability(desc1)
        scores = sigmoid_log_double_softmax(sim, z0, z1)
        return scores, sim

    def scores(self, desc0: torch.Tensor, desc1: torch.Tensor):
        m0 = torch.sigmoid(self.matchability(desc0)).squeeze(-1)
        m1 = torch.sigmoid(self.matchability(desc1)).squeeze(-1)
        return m0, m1


def filter_matches(scores: torch.Tensor, th: float):
    """ obtain matches from a log assignment matrix [Bx M+1 x N+1]"""
    max0, max1 = scores[:, :-1, :-1].max(2), scores[:, :-1, :-1].max(1)
    m0, m1 = max0.indices, max1.indices
    mutual0 = torch.arange(m0.shape[1]).to(m0)[None] == m1.gather(1, m0)
    mutual1 = torch.arange(m1.shape[1]).to(m1)[None] == m0.gather(1, m1)
    max0_exp = max0.values.exp()
    zero = max0_exp.new_tensor(0)
    mscores0 = torch.where(mutual0, max0_exp, zero)
    mscores1 = torch.where(mutual1, mscores0.gather(1, m1), zero)
    if th is not None:
        valid0 = mutual0 & (mscores0 > th)
    else:
        valid0 = mutual0
    valid1 = mutual1 & valid0.gather(1, m1)
    m0 = torch.where(valid0, m0, m0.new_tensor(-1))
    m1 = torch.where(valid1, m1, m1.new_tensor(-1))
    return m0, m1, mscores0, mscores1


class LightGlue(nn.Module):
    default_conf = {
        "name": "lightglue",  # just for interfacing
        "input_dim": 256,  # input descriptor dimension (autoselected from weights)
        "descriptor_dim": 256,
        "n_layers": 9,
        "num_heads": 4,
        "flash": True,  # enable FlashAttention if available.
        "mp": False,  # enable mixed precision
        "depth_confidence": 0.95,  # early stopping, disable with -1
        "width_confidence": 0.99,  # point pruning, disable with -1
        "filter_threshold": 0.1,  # match threshold
        "weights": None,
    }

    required_data_keys = ["image0", "image1"]

    version = "v0.1_arxiv"
    url = "https://github.com/cvg/LightGlue/releases/download/{}/{}_lightglue.pth"

    features = {
        "superpoint": ("superpoint_lightglue", 256),
        "disk": ("disk_lightglue", 128),
    }

    def __init__(self, features="superpoint", **conf) -> None:
        super().__init__()
        self.conf = {**self.default_conf, **conf}
        if features is not None:
            assert features in list(self.features.keys())
            self.conf["weights"], self.conf["input_dim"] = self.features[features]
        self.conf = conf = SimpleNamespace(**self.conf)

        if conf.input_dim != conf.descriptor_dim:
            self.input_proj = nn.Linear(conf.input_dim, conf.descriptor_dim, bias=True)
        else:
            self.input_proj = nn.Identity()

        head_dim = conf.descriptor_dim // conf.num_heads
        self.posenc = LearnableFourierPositionalEncoding(2, head_dim, head_dim)

        h, n, d = conf.num_heads, conf.n_layers, conf.descriptor_dim
        self.self_attn = nn.ModuleList(
            [Transformer(d, h, conf.flash) for _ in range(n)]
        )
        self.cross_attn = nn.ModuleList(
            [CrossTransformer(d, h, conf.flash) for _ in range(n)]
        )
        self.log_assignment = nn.ModuleList([MatchAssignment(d) for _ in range(n)])
        self.token_confidence = nn.ModuleList(
            [TokenConfidence(d) for _ in range(n - 1)]
        )

        if features is not None:
            fname = f"{conf.weights}_{self.version}.pth".replace(".", "-")
            state_dict = torch.hub.load_state_dict_from_url(
                self.url.format(self.version, features), file_name=fname
            )
            self.load_state_dict(state_dict, strict=False)
        elif conf.weights is not None:
            path = Path(__file__).parent
            path = path / "weights/{}.pth".format(self.conf.weights)
            state_dict = torch.load(str(path), map_location="cpu")
            self.load_state_dict(state_dict, strict=False)

        print("Loaded LightGlue model")

    def forward(self, data: dict) -> dict:
        """
        Match keypoints and descriptors between two images

        Input (dict):
            image0: dict
                keypoints: [B x M x 2]
                descriptors: [B x M x D]
                image: [B x C x H x W] or image_size: [B x 2]
            image1: dict
                keypoints: [B x N x 2]
                descriptors: [B x N x D]
                image: [B x C x H x W] or image_size: [B x 2]
        Output (dict):
            log_assignment: [B x M+1 x N+1]
            matches0: [B x M]
            matching_scores0: [B x M]
            matches1: [B x N]
            matching_scores1: [B x N]
            matches: List[[Si x 2]], scores: List[[Si]]
        """
        with torch.autocast(enabled=self.conf.mp, device_type="cuda"):
            return self._forward(data)

    def _forward(self, data: dict) -> dict:
        for key in self.required_data_keys:
            assert key in data, f"Missing key {key} in data"
        data0, data1 = data["image0"], data["image1"]
        kpts0_, kpts1_ = data0["keypoints"], data1["keypoints"]
        b, m, _ = kpts0_.shape
        b, n, _ = kpts1_.shape
        size0, size1 = data0.get("image_size"), data1.get("image_size")
        size0 = size0 if size0 is not None else data0["image"].shape[-2:][::-1]
        size1 = size1 if size1 is not None else data1["image"].shape[-2:][::-1]
        kpts0 = normalize_keypoints(kpts0_, size=size0)
        kpts1 = normalize_keypoints(kpts1_, size=size1)

        assert torch.all(kpts0 >= -1) and torch.all(kpts0 <= 1)
        assert torch.all(kpts1 >= -1) and torch.all(kpts1 <= 1)

        desc0 = data0["descriptors"].detach()
        desc1 = data1["descriptors"].detach()

        assert desc0.shape[-1] == self.conf.input_dim
        assert desc1.shape[-1] == self.conf.input_dim

        if torch.is_autocast_enabled():
            desc0 = desc0.half()
            desc1 = desc1.half()

        desc0 = self.input_proj(desc0)
        desc1 = self.input_proj(desc1)

        # cache positional embeddings
        encoding0 = self.posenc(kpts0)
        encoding1 = self.posenc(kpts1)

        # GNN + final_proj + assignment
        ind0 = torch.arange(0, m).to(device=kpts0.device)[None]
        ind1 = torch.arange(0, n).to(device=kpts0.device)[None]
        prune0 = torch.ones_like(ind0)  # store layer where pruning is detected
        prune1 = torch.ones_like(ind1)
        dec, wic = self.conf.depth_confidence, self.conf.width_confidence
        token0, token1 = None, None
        for i in range(self.conf.n_layers):
            # self+cross attention
            desc0, desc1 = self.self_attn[i](desc0, desc1, encoding0, encoding1)
            desc0, desc1 = self.cross_attn[i](desc0, desc1)
            if i == self.conf.n_layers - 1:
                continue  # no early stopping or adaptive width at last layer
            if dec > 0:  # early stopping
                token0, token1 = self.token_confidence[i](desc0, desc1)
                if self.stop(token0, token1, self.conf_th(i), dec, m + n):
                    break
            if wic > 0:  # point pruning
                match0, match1 = self.log_assignment[i].scores(desc0, desc1)
                mask0 = self.get_mask(token0, match0, self.conf_th(i), 1 - wic)
                mask1 = self.get_mask(token1, match1, self.conf_th(i), 1 - wic)
                ind0, ind1 = ind0[mask0][None], ind1[mask1][None]
                desc0, desc1 = desc0[mask0][None], desc1[mask1][None]
                if desc0.shape[-2] == 0 or desc1.shape[-2] == 0:
                    break
                encoding0 = encoding0[:, :, mask0][:, None]
                encoding1 = encoding1[:, :, mask1][:, None]
            prune0[:, ind0] += 1
            prune1[:, ind1] += 1

        if wic > 0:  # scatter with indices after pruning
            scores_, _ = self.log_assignment[i](desc0, desc1)
            dt, dev = scores_.dtype, scores_.device
            scores = torch.zeros(b, m + 1, n + 1, dtype=dt, device=dev)
            scores[:, :-1, :-1] = -torch.inf
            scores[:, ind0[0], -1] = scores_[:, :-1, -1]
            scores[:, -1, ind1[0]] = scores_[:, -1, :-1]
            x, y = torch.meshgrid(ind0[0], ind1[0], indexing="ij")
            scores[:, x, y] = scores_[:, :-1, :-1]
        else:
            scores, _ = self.log_assignment[i](desc0, desc1)

        m0, m1, mscores0, mscores1 = filter_matches(scores, self.conf.filter_threshold)

        matches, mscores = [], []
        for k in range(b):
            valid = m0[k] > -1
            matches.append(torch.stack([torch.where(valid)[0], m0[k][valid]], -1))
            mscores.append(mscores0[k][valid])

        return {
            "log_assignment": scores,
            "matches0": m0,
            "matches1": m1,
            "matching_scores0": mscores0,
            "matching_scores1": mscores1,
            "stop": i + 1,
            "prune0": prune0,
            "prune1": prune1,
            "matches": matches,
            "scores": mscores,
        }

    def conf_th(self, i: int) -> float:
        """ scaled confidence threshold """
        return np.clip(0.8 + 0.1 * np.exp(-4.0 * i / self.conf.n_layers), 0, 1)

    def get_mask(
        self,
        confidence: torch.Tensor,
        match: torch.Tensor,
        conf_th: float,
        match_th: float,
    ) -> torch.Tensor:
        """ mask points which should be removed """
        if conf_th and confidence is not None:
            mask = (
                torch.where(confidence > conf_th, match, match.new_tensor(1.0))
                > match_th
            )
        else:
            mask = match > match_th
        return mask

    def stop(
        self,
        token0: torch.Tensor,
        token1: torch.Tensor,
        conf_th: float,
        inl_th: float,
        seql: int,
    ) -> torch.Tensor:
        """ evaluate stopping condition"""
        tokens = torch.cat([token0, token1], -1)
        if conf_th:
            pos = 1.0 - (tokens < conf_th).float().sum() / seql
            return pos > inl_th
        else:
            return tokens.mean() > inl_th



from types import SimpleNamespace

import kornia
import torch
import torch.nn as nn

from .utils import ImagePreprocessor


class DISK(nn.Module):
    default_conf = {
        "weights": "depth",
        "max_num_keypoints": None,
        "desc_dim": 128,
        "nms_window_size": 5,
        "detection_threshold": 0.0,
        "pad_if_not_divisible": True,
    }

    preprocess_conf = {
        **ImagePreprocessor.default_conf,
        "resize": 1024,
        "grayscale": False,
    }

    required_data_keys = ["image"]

    def __init__(self, **conf) -> None:
        super().__init__()
        self.conf = {**self.default_conf, **conf}
        self.conf = SimpleNamespace(**self.conf)
        self.model = kornia.feature.DISK.from_pretrained(self.conf.weights)

    def forward(self, data: dict) -> dict:
        """ Compute keypoints, scores, descriptors for image """
        for key in self.required_data_keys:
            assert key in data, f"Missing key {key} in data"
        image = data["image"]
        features = self.model(
            image,
            n=self.conf.max_num_keypoints,
            window_size=self.conf.nms_window_size,
            score_threshold=self.conf.detection_threshold,
            pad_if_not_divisible=self.conf.pad_if_not_divisible,
        )
        keypoints = [f.keypoints for f in features]
        scores = [f.detection_scores for f in features]
        descriptors = [f.descriptors for f in features]
        del features

        keypoints = torch.stack(keypoints, 0)
        scores = torch.stack(scores, 0)
        descriptors = torch.stack(descriptors, 0)

        return {
            "keypoints": keypoints.to(image),
            "keypoint_scores": scores.to(image),
            "descriptors": descriptors.to(image),
        }

    def extract(self, img: torch.Tensor, **conf) -> dict:
        """ Perform extraction with online resizing"""
        if img.dim() == 3:
            img = img[None]  # add batch dim
        assert img.dim() == 4 and img.shape[0] == 1
        shape = img.shape[-2:][::-1]
        img, scales = ImagePreprocessor(**{**self.preprocess_conf, **conf})(img)
        feats = self.forward({"image": img})
        feats["image_size"] = torch.tensor(shape)[None].to(img).float()
        feats["keypoints"] = (feats["keypoints"] + 0.5) / scales[None] - 0.5
        return feats



# %BANNER_BEGIN%
# ---------------------------------------------------------------------
# %COPYRIGHT_BEGIN%
#
#  Magic Leap, Inc. ("COMPANY") CONFIDENTIAL
#
#  Unpublished Copyright (c) 2020
#  Magic Leap, Inc., All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property
# of COMPANY. The intellectual and technical concepts contained herein
# are proprietary to COMPANY and may be covered by U.S. and Foreign
# Patents, patents in process, and are protected by trade secret or
# copyright law.  Dissemination of this information or reproduction of
# this material is strictly forbidden unless prior written permission is
# obtained from COMPANY.  Access to the source code contained herein is
# hereby forbidden to anyone except current COMPANY employees, managers
# or contractors who have executed Confidentiality and Non-disclosure
# agreements explicitly covering such access.
#
# The copyright notice above does not evidence any actual or intended
# publication or disclosure  of  this source code, which includes
# information that is confidential and/or proprietary, and is a trade
# secret, of  COMPANY.   ANY REPRODUCTION, MODIFICATION, DISTRIBUTION,
# PUBLIC  PERFORMANCE, OR PUBLIC DISPLAY OF OR THROUGH USE  OF THIS
# SOURCE CODE  WITHOUT THE EXPRESS WRITTEN CONSENT OF COMPANY IS
# STRICTLY PROHIBITED, AND IN VIOLATION OF APPLICABLE LAWS AND
# INTERNATIONAL TREATIES.  THE RECEIPT OR POSSESSION OF  THIS SOURCE
# CODE AND/OR RELATED INFORMATION DOES NOT CONVEY OR IMPLY ANY RIGHTS
# TO REPRODUCE, DISCLOSE OR DISTRIBUTE ITS CONTENTS, OR TO MANUFACTURE,
# USE, OR SELL ANYTHING THAT IT  MAY DESCRIBE, IN WHOLE OR IN PART.
#
# %COPYRIGHT_END%
# ----------------------------------------------------------------------
# %AUTHORS_BEGIN%
#
#  Originating Authors: Paul-Edouard Sarlin
#
# %AUTHORS_END%
# --------------------------------------------------------------------*/
# %BANNER_END%

# Adapted by Remi Pautrat, Philipp Lindenberger

import torch
from torch import nn

from .utils import ImagePreprocessor


def simple_nms(scores, nms_radius: int):
    """ Fast Non-maximum suppression to remove nearby points """
    assert nms_radius >= 0

    def max_pool(x):
        return torch.nn.functional.max_pool2d(
            x, kernel_size=nms_radius * 2 + 1, stride=1, padding=nms_radius
        )

    zeros = torch.zeros_like(scores)
    max_mask = scores == max_pool(scores)
    for _ in range(2):
        supp_mask = max_pool(max_mask.float()) > 0
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == max_pool(supp_scores)
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)


def top_k_keypoints(keypoints, scores, k):
    if k >= len(keypoints):
        return keypoints, scores
    scores, indices = torch.topk(scores, k, dim=0, sorted=True)
    return keypoints[indices], scores


def sample_descriptors(keypoints, descriptors, s: int = 8):
    """ Interpolate descriptors at keypoint locations """
    b, c, h, w = descriptors.shape
    keypoints = keypoints - s / 2 + 0.5
    keypoints /= torch.tensor([(w * s - s / 2 - 0.5), (h * s - s / 2 - 0.5)]).to(
        keypoints
    )[None]
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
    args = {"align_corners": True} if torch.__version__ >= "1.3" else {}
    descriptors = torch.nn.functional.grid_sample(
        descriptors, keypoints.view(b, 1, -1, 2), mode="bilinear", **args
    )
    descriptors = torch.nn.functional.normalize(
        descriptors.reshape(b, c, -1), p=2, dim=1
    )
    return descriptors


class SuperPoint(nn.Module):
    """SuperPoint Convolutional Detector and Descriptor

    SuperPoint: Self-Supervised Interest Point Detection and
    Description. Daniel DeTone, Tomasz Malisiewicz, and Andrew
    Rabinovich. In CVPRW, 2019. https://arxiv.org/abs/1712.07629

    """

    default_conf = {
        "descriptor_dim": 256,
        "nms_radius": 4,
        "max_num_keypoints": None,
        "detection_threshold": 0.0005,
        "remove_borders": 4,
    }

    preprocess_conf = {
        **ImagePreprocessor.default_conf,
        "resize": 1024,
        "grayscale": True,
    }

    required_data_keys = ["image"]

    def __init__(self, **conf):
        super().__init__()
        self.conf = {**self.default_conf, **conf}

        self.relu = nn.ReLU(inplace=True)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        c1, c2, c3, c4, c5 = 64, 64, 128, 128, 256

        self.conv1a = nn.Conv2d(1, c1, kernel_size=3, stride=1, padding=1)
        self.conv1b = nn.Conv2d(c1, c1, kernel_size=3, stride=1, padding=1)
        self.conv2a = nn.Conv2d(c1, c2, kernel_size=3, stride=1, padding=1)
        self.conv2b = nn.Conv2d(c2, c2, kernel_size=3, stride=1, padding=1)
        self.conv3a = nn.Conv2d(c2, c3, kernel_size=3, stride=1, padding=1)
        self.conv3b = nn.Conv2d(c3, c3, kernel_size=3, stride=1, padding=1)
        self.conv4a = nn.Conv2d(c3, c4, kernel_size=3, stride=1, padding=1)
        self.conv4b = nn.Conv2d(c4, c4, kernel_size=3, stride=1, padding=1)

        self.convPa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convPb = nn.Conv2d(c5, 65, kernel_size=1, stride=1, padding=0)

        self.convDa = nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convDb = nn.Conv2d(
            c5, self.conf["descriptor_dim"], kernel_size=1, stride=1, padding=0
        )

        url = "https://github.com/cvg/LightGlue/releases/download/v0.1_arxiv/superpoint_v1.pth"
        self.load_state_dict(torch.hub.load_state_dict_from_url(url))

        mk = self.conf["max_num_keypoints"]
        if mk is not None and mk <= 0:
            raise ValueError("max_num_keypoints must be positive or None")

        print("Loaded SuperPoint model")

    def forward(self, data: dict) -> dict:
        """ Compute keypoints, scores, descriptors for image """
        for key in self.required_data_keys:
            assert key in data, f"Missing key {key} in data"
        image = data["image"]
        if image.shape[1] == 3:  # RGB
            scale = image.new_tensor([0.299, 0.587, 0.114]).view(1, 3, 1, 1)
            image = (image * scale).sum(1, keepdim=True)
        # Shared Encoder
        x = self.relu(self.conv1a(image))
        x = self.relu(self.conv1b(x))
        x = self.pool(x)
        x = self.relu(self.conv2a(x))
        x = self.relu(self.conv2b(x))
        x = self.pool(x)
        x = self.relu(self.conv3a(x))
        x = self.relu(self.conv3b(x))
        x = self.pool(x)
        x = self.relu(self.conv4a(x))
        x = self.relu(self.conv4b(x))

        # Compute the dense keypoint scores
        cPa = self.relu(self.convPa(x))
        scores = self.convPb(cPa)
        scores = torch.nn.functional.softmax(scores, 1)[:, :-1]
        b, _, h, w = scores.shape
        scores = scores.permute(0, 2, 3, 1).reshape(b, h, w, 8, 8)
        scores = scores.permute(0, 1, 3, 2, 4).reshape(b, h * 8, w * 8)
        scores = simple_nms(scores, self.conf["nms_radius"])

        # Discard keypoints near the image borders
        if self.conf["remove_borders"]:
            pad = self.conf["remove_borders"]
            scores[:, :pad] = -1
            scores[:, :, :pad] = -1
            scores[:, -pad:] = -1
            scores[:, :, -pad:] = -1

        # Extract keypoints
        best_kp = torch.where(scores > self.conf["detection_threshold"])
        scores = scores[best_kp]

        # Separate into batches
        keypoints = [
            torch.stack(best_kp[1:3], dim=-1)[best_kp[0] == i] for i in range(b)
        ]
        scores = [scores[best_kp[0] == i] for i in range(b)]

        # Keep the k keypoints with highest score
        if self.conf["max_num_keypoints"] is not None:
            keypoints, scores = list(
                zip(
                    *[
                        top_k_keypoints(k, s, self.conf["max_num_keypoints"])
                        for k, s in zip(keypoints, scores)
                    ]
                )
            )

        # Convert (h, w) to (x, y)
        keypoints = [torch.flip(k, [1]).float() for k in keypoints]

        # Compute the dense descriptors
        cDa = self.relu(self.convDa(x))
        descriptors = self.convDb(cDa)
        descriptors = torch.nn.functional.normalize(descriptors, p=2, dim=1)

        # Extract descriptors
        descriptors = [
            sample_descriptors(k[None], d[None], 8)[0]
            for k, d in zip(keypoints, descriptors)
        ]

        return {
            "keypoints": torch.stack(keypoints, 0),
            "keypoint_scores": torch.stack(scores, 0),
            "descriptors": torch.stack(descriptors, 0).transpose(-1, -2),
        }

    def extract(self, img: torch.Tensor, **conf) -> dict:
        """ Perform extraction with online resizing"""
        if img.dim() == 3:
            img = img[None]  # add batch dim
        assert img.dim() == 4 and img.shape[0] == 1
        shape = img.shape[-2:][::-1]
        img, scales = ImagePreprocessor(**{**self.preprocess_conf, **conf})(img)
        feats = self.forward({"image": img})
        feats["image_size"] = torch.tensor(shape)[None].to(img).float()
        feats["keypoints"] = (feats["keypoints"] + 0.5) / scales[None] - 0.5
        return feats



from abc import ABC, abstractmethod
from typing import List, Optional, Union
import torch
from opensr_test.config import Metric
from opensr_test.utils import check_lpips, check_openclip
from torchvision.transforms import Normalize

class DistanceMetric(ABC):
    """ Metaclass to define the distance metrics

    Parameters:
        method (str): The method to use. Either "pixel", "patch", or "image".
        patch_size (int): The patch size to use if the patch method is used.
        x (torch.Tensor): The SR harmonized image (C, H, W).
        y (torch.Tensor): The HR image (C, H, W).
        **kwargs: The parameters to pass to the distance function.
        
    Abstract methods:
        compute_patch: Compute the distance metric at patch level.
        compute_image: Compute the distance metric at image level.
        compute_pixel: Compute the distance metric at image level.
        compute: Compute the distance metric.
        
    """

    def __init__(
        self,
        name: str,
        method: str,
        patch_size: int,
        x: torch.Tensor,
        y: torch.Tensor,
        **kwargs
    ):
        self.method = method
        self.name = name
        self.patch_size = patch_size
        self.kwargs = kwargs
        self.axis: int = 0
        self.x = x
        self.y = y

    @staticmethod
    def do_square(tensor: torch.Tensor, patch_size: Optional[int] = 32) -> torch.Tensor:
        """ Split a tensor into n_patches x n_patches patches and return
        the patches as a tensor.
        
        Args:
            tensor (torch.Tensor): The tensor to split.
            n_patches (int, optional): The number of patches to split the tensor into.
                If None, the tensor is split into the smallest number of patches.

        Returns:
            torch.Tensor: The patches as a tensor.
        """
        # Check if it is a square tensor
        if tensor.shape[-1] != tensor.shape[-2]:
            raise ValueError("The tensor must be square.")

        # tensor (C, H, W)
        xdim = tensor.shape[1]
        ydim = tensor.shape[2]

        minimages_x = int(torch.ceil(torch.tensor(xdim / patch_size)))
        minimages_y = int(torch.ceil(torch.tensor(ydim / patch_size)))

        pad_x_01 = int((minimages_x * patch_size - xdim) // 2)
        pad_x_02 = int((minimages_x * patch_size - xdim) - pad_x_01)

        pad_y_01 = int((minimages_y * patch_size - ydim) // 2)
        pad_y_02 = int((minimages_y * patch_size - ydim) - pad_y_01)

        padded_tensor = torch.nn.functional.pad(
            tensor, (pad_x_01, pad_x_02, pad_y_01, pad_y_02)
        )

        # split the tensor (C, H, W) into (n_patches, n_patches, C, H, W)
        patches = padded_tensor.unfold(1, patch_size, patch_size).unfold(
            2, patch_size, patch_size
        )

        # move the axes (C, n_patches, n_patches, H, W) -> (n_patches, n_patches, C, H, W)
        patches = patches.permute(1, 2, 0, 3, 4)

        return patches

    @abstractmethod
    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        pass

    def compute_image(self) -> torch.Tensor:
        return Metric(value=self._compute_image(self.x, self.y), description=self.name)

    def compute_patch(self) -> torch.Tensor:
        x_batched = self.do_square(self.x, self.patch_size)
        y_batched = self.do_square(self.y, self.patch_size)
        metric_result = torch.zeros(x_batched.shape[:2])

        xrange, yrange = x_batched.shape[0:2]

        for x_index in range(xrange):
            for y_index in range(yrange):
                x_batch = x_batched[x_index, y_index]
                y_batch = y_batched[x_index, y_index]
                metric_result[x_index, y_index] = self._compute_image(x_batch, y_batch)

        # Go back to the original size
        metric_result = torch.nn.functional.interpolate(
            metric_result[None, None], size=self.x.shape[-2:], mode="nearest"
        ).squeeze()
        
        return Metric(value=metric_result, description=self.name)

    def compute_pixel(self) -> torch.Tensor:
        return Metric(value=self._compute_pixel(self.x, self.y), description=self.name)

    def compute(self) -> torch.Tensor:
        if self.method == "pixel":
            return self.compute_pixel()
        elif self.method == "image":
            return self.compute_image()
        elif self.method == "patch":
            return self.compute_patch()
        else:
            raise ValueError("Invalid method.")


class KL(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="KL")

        self.large_number = 1e2
        self.epsilon = 1e-8

        pbias_d = torch.abs(x / (y + self.epsilon))
        pbias_d[pbias_d > self.large_number] = self.large_number
        pbias_d[pbias_d < 1 / self.large_number] = 1 / self.large_number

        self.pbias_d = pbias_d

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.mean(x) * torch.log(torch.mean(self.pbias_d))

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.mean(x, axis=0) * torch.log(torch.mean(self.pbias_d, axis=0))


class L1(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="L1")

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean(torch.abs(x - y))

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean(torch.abs(x - y), axis=0)


class L2(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="L2")

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean((x - y) ** 2)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean((x - y) ** 2, axis=0)


class PBIAS(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="PBIAS")

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean((x - y) ** 2)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.nanmean((x - y) ** 2, axis=0)


class PSNR(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="PSNR")

        self.data_range = torch.tensor(1)
        self.epsilon = torch.tensor(1e-10)

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        l2_distance = torch.nanmean((x - y) ** 2)
        l2_distance[l2_distance < self.epsilon] = self.epsilon
        return 10 * torch.log10(self.data_range ** 2 / l2_distance)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        l2_distance = torch.nanmean((x - y) ** 2, axis=0)
        l2_distance[l2_distance < self.epsilon] = self.epsilon
        return 10 * torch.log10(self.data_range ** 2 / l2_distance)


class SAD(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="SAD")

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dot_product = (x * y).sum(dim=0)
        preds_norm = x.norm(dim=0)
        target_norm = y.norm(dim=0)
        sam_score = torch.clamp(dot_product / (preds_norm * target_norm), -1, 1).acos()
        return torch.rad2deg(sam_score)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dot_product = (x * y).squeeze().sum()
        preds_norm = x.squeeze().norm()
        target_norm = y.squeeze().norm()
        sam_score = torch.clamp(dot_product / (preds_norm * target_norm), -1, 1).acos()
        return torch.rad2deg(sam_score)


class SAD(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
    ):
        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="SAD")

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dot_product = (x * y).sum(dim=0)
        preds_norm = x.norm(dim=0)
        target_norm = y.norm(dim=0)
        sam_score = torch.clamp(dot_product / (preds_norm * target_norm), -1, 1).acos()
        return torch.rad2deg(sam_score)

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dot_product = (x * y).sum()
        preds_norm = x.norm()
        target_norm = y.norm()
        sam_score = torch.clamp(dot_product / (preds_norm * target_norm), -1, 1).acos()
        return torch.rad2deg(sam_score)


class LPIPS(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
        device: Union[str, torch.device] = "cpu",
    ):
        # if extra_requires (setup.py) is not fulfilled, raise error
        check_lpips()
        import lpips

        if method == "patch":
            if patch_size < 32:
                raise ValueError("The patch size must be at least 32.")

        # Set the model
        self.model = lpips.LPIPS(net="alex", verbose=False).to(device)

        # Normalize the tensors to [-1, 1]
        y = (y - y.min()) / (y.max() - y.min())
        y = y * 2 - 1

        x = (x - x.min()) / (x.max() - x.min())
        x = x * 2 - 1

        super().__init__(x=x, y=y, method=method, patch_size=patch_size, name="LPIPS")

    @torch.no_grad()
    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.model(x, y).mean()

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("LPIPS cannot be computed at pixel level.")


class CLIPscore(DistanceMetric):
    """Estimate the CLIPscore between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
        device: Union[str, torch.device] = "cpu",
    ):
        # if extra_requires (setup.py) is not fulfilled, raise error
        check_openclip()
        import open_clip

        if method != "patch":
            raise ValueError(
                "CLIPscore can only be computed at patch level of 256x256."
            )

        if method == "patch":
            if patch_size != 256:
                raise ValueError("The patch size must be 256.")

        # Set the model        
        self.model = open_clip.create_model_and_transforms(
            "ViT-B-16-SigLIP-256",
            pretrained='webli',
            device=device
        )[0]
        
        # Normalize the tensors
        normalize = Normalize(
            mean=(0.48145466, 0.4578275, 0.40821073),
            std=(0.26862954, 0.26130258, 0.27577711)
        )
        x_norm = normalize(x)
        y_norm = normalize(y)

        super().__init__(
            x=x_norm, y=y_norm, method=method, patch_size=patch_size, name="CLIPscore"
        )

    @torch.no_grad()
    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x_emb = self.model.encode_image(x[None]).squeeze()
        y_emb = self.model.encode_image(y[None]).squeeze()
        return torch.nn.functional.l1_loss(x_emb, y_emb)

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("CLIPscore cannot be computed at pixel level.")



class CROSSMTF(DistanceMetric):
    """Spectral information divergence between two tensors"""

    def __init__(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        method: str = "image",
        patch_size: int = 32,
        scale: int = 4,
    ):
        super().__init__(
            x=x, y=y, method=method, patch_size=patch_size, name="CROSSMTF"
        )

        if method == "patch":
            if patch_size < 32:
                raise ValueError("The patch size must be at least 32.")

        self.scale = scale

    def _compute_image(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # Compute mask
        freq = torch.fft.fftfreq(x.shape[-1])
        freq = torch.fft.fftshift(freq)
        kfreq2D = torch.meshgrid(freq, freq, indexing="ij")
        knrm = torch.sqrt(kfreq2D[0] ** 2 + kfreq2D[1] ** 2)
        tensormask = (knrm > (0.5 * 1 / self.scale)) & (knrm < (0.5))

        fft_preds = torch.abs(torch.fft.fftshift(torch.fft.fft2(x)))
        fft_target = torch.abs(torch.fft.fftshift(torch.fft.fft2(y)))

        mtf = torch.masked_select(fft_preds / fft_target, tensormask)

        return torch.mean(torch.clamp(mtf, 0, 1))

    def _compute_pixel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("CROSSMTF cannot be computed at pixel level.")


def get_distance(
    x: torch.Tensor,
    y: torch.Tensor,
    method: str,
    agg_method: str,
    patch_size: int = 32,
    scale: int = 4,
    device: Union[str, torch.device] = "cpu",
    rgb_bands: Optional[List[int]] = [0, 1, 2],
):
    """ Estimate the unsystematic error between the SR and HR image.

    Args:
        x (torch.Tensor): The SR harmonized image (C, H, W).
        hr (torch.Tensor): The HR image (C, H, W).
        method (str): The method to use. Either "psnr" or "cpsnr".
        agg_method (str): The method to use to aggregate the distance.
            Either "pixel", "image", or "patch".
        patch_size (int, optional): The patch size to use if the patch 
            method is used.
        scale (int, optional): The scale of the super-resolution.
        space_search (int, optional): This parameter is used to search 
            for the best shift that maximizes the PSNR. By default, it is 
            the same as the super-resolution scale.
        
    Returns:
        torch.Tensor: The metric value.
    """
    if x.shape[0] != y.shape[0]:
        raise ValueError("The number of channels in x and y must be the same.")

    if x.shape[1] != y.shape[1]:
        raise ValueError("The height of x and y must be the same.")

    if method == "kl":
        distance_fn = KL(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "l1":
        distance_fn = L1(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "l2":
        distance_fn = L2(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "pbias":
        distance_fn = PBIAS(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "ipsnr":
        distance_fn = 1 / PSNR(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "sad":
        distance_fn = SAD(x=x, y=y, method=agg_method, patch_size=patch_size)
    elif method == "crossmtf":
        distance_fn = CROSSMTF(
            x=x, y=y, method=agg_method, patch_size=patch_size, scale=scale
        )
    elif method == "lpips":
        x_rgb = x[rgb_bands, :, :]
        y_rgb = y[rgb_bands, :, :]
        distance_fn = LPIPS(
            x_rgb, y_rgb, method=agg_method, patch_size=patch_size, device=device
        )
    elif method == "clip":
        x_rgb = x[rgb_bands, :, :]
        y_rgb = y[rgb_bands, :, :]
        distance_fn = CLIPscore(
            x_rgb, y_rgb, method=agg_method, patch_size=patch_size, device=device
        )
    else:
        raise ValueError("No valid distance method.")

    return distance_fn



import pathlib
from typing import Optional
from opensr_test.config import create_param_config
import numpy as np
import requests
import torch


def download(url: str, save_path: str) -> str:
    """ Download a file from a url.

    Args:
        url (str): The url of the file in HuggingFace Hub.
        save_path (str): The path to save the file.

    Returns:
        str: The path to the file.
    """
    response = requests.get(url, stream=True)
    with open(save_path, "wb") as f:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    return None

def load(
    dataset: str = "naip", model_dir: Optional[str] = None, force: bool = False
) -> torch.Tensor:
    """ Load a dataset.

    Args:
        dataset (str, optional): The dataset to load. Defaults to "naip".
        force (bool, optional): If True, force the download. Defaults to False.
        
    Raises:
        NotImplementedError: If the dataset is not implemented.

    Returns:
        torch.Tensor: The dataset in a tensor of shape (N, C, H, W).
    """
    if model_dir is None:
        ROOT_FOLDER = get_data_path()
    else:
        ROOT_FOLDER = pathlib.Path(model_dir)

    DATASETS = ["naip", "spot", "venus"]
    if dataset not in DATASETS:
        raise NotImplementedError("The dataset %s is not implemented." % dataset)

    URL = "https://huggingface.co/csaybar/opensr-test/resolve/main"

    # Create folder
    [(ROOT_FOLDER / x).mkdir(exist_ok=True) for x in DATASETS]

    # Download the files

    ## hr file
    hrfile_url = "%s/%s/%s" % (URL, dataset, "hr.npy")
    hrfile_path = ROOT_FOLDER / dataset / "hr.npy"
    if not hrfile_path.exists() or force:
        print(f"Downloading {dataset} dataset  - {hrfile_path.stem} file.")
        download(hrfile_url, hrfile_path)

    ## lr file
    lrfile_url = "%s/%s/lr.npy" % (URL, dataset)
    lrfile_path = ROOT_FOLDER / dataset / "lr.npy"
    if not lrfile_path.exists() or force:
        print(f"Downloading {dataset} dataset  - {lrfile_path.stem} file.")
        download(lrfile_url, lrfile_path)

    ## landuse file
    landuse_url = "%s/%s/landuse2.npy" % (URL, dataset)
    landuse_path = ROOT_FOLDER / dataset / "landuse2.npy"
    if not landuse_path.exists() or force:
        print(f"Downloading {dataset} dataset  - {landuse_path.stem} file.")
        download(landuse_url, landuse_path)

    ## metadata file
    csvfile_url = "%s/%s/metadata.csv" % (URL, dataset)
    csvfile_path = ROOT_FOLDER / dataset / "metadata.csv"
    if not csvfile_path.exists() or force:
        print(f"Downloading {dataset} dataset  - {csvfile_path.stem} file.")
        download(csvfile_url, csvfile_path)

    # Load the dataset

    ## LR file
    lr_data = np.load(ROOT_FOLDER / dataset / "lr.npy") / 10000
    lr_data_torch = lr_data.astype(np.float32)

    ## HR file
    hr_data = np.load(ROOT_FOLDER / dataset / "hr.npy") / 10000
    hr_data_torch = hr_data.astype(np.float32)

    ## LandUse file
    land_use = np.load(ROOT_FOLDER / dataset / "landuse2.npy")
    land_use_torch = land_use

    return {
        "lr": lr_data_torch,
        "hr": hr_data_torch,
        "landuse": land_use_torch,
        "params": create_param_config(dataset)
    }


def get_data_path() -> str:
    cred_path = pathlib.Path.home() / ".config/opensr_test/"
    cred_path.mkdir(parents=True, exist_ok=True)
    return cred_path



from typing import List, Optional

import torch
import numpy as np
from opensr_test.distance import get_distance

def sigm(x):
    return 1 / (1 + np.exp(-(x - 0) / 1)) 

def get_distances(
    lr_to_hr: torch.Tensor,
    sr_harm: torch.Tensor,
    hr: torch.Tensor,
    distance_method: str = "l1",
    agg_method: str = "pixel",
    scale: int = 4,
    patch_size: Optional[int] = 32,
    rgb_bands: Optional[List[int]] = [0, 1, 2],
    device: str = "cpu",
):
    """ Obtain the distances between metrics in a 2D space.
        In the 2D space the y-axis is the distance between the SR and HR
        (error) and the x-axis is the distance between the LR and HR 
        (high-frequency).
        

    Args:
        lr_to_hr (torch.Tensor): The LR image upscaled to HR.
        sr_harm (torch.Tensor): The SR image harmonized to HR.
        hr (torch.Tensor): The HR image.
        distance_method (str, optional): The distance method to use. Defaults to 
            "psnr". Available methods are: psnr, kl, l1, l2, psnr, sad, lpips.

    Returns:
        torch.Tensor: The distances between the metrics.
    """

    reference = get_distance(
        x=lr_to_hr,
        y=hr,
        method=distance_method,
        agg_method=agg_method,
        patch_size=patch_size,
        device=device,
        scale=scale,
        rgb_bands=rgb_bands,
    ).compute()

    dist_sr_to_hr = get_distance(
        x=sr_harm,
        y=hr,
        method=distance_method,
        agg_method=agg_method,
        patch_size=patch_size,
        device=device,
        scale=scale,
        rgb_bands=rgb_bands,
    ).compute()

    dist_sr_to_lr = get_distance(
        x=sr_harm,
        y=lr_to_hr,
        method=distance_method,
        agg_method=agg_method,
        patch_size=patch_size,
        device=device,
        scale=scale,
        rgb_bands=rgb_bands,
    ).compute()

    return reference.value, dist_sr_to_hr.value, dist_sr_to_lr.value


def tc_improvement(
    d_im: torch.Tensor,
    d_om: torch.Tensor,
    plambda: float = 0.85
) -> torch.Tensor:
    """ Obtain the relative distance to the center
    of the improvement space

    Args:
        d_im (torch.Tensor): The distance to the improvement space
        d_om (torch.Tensor): The distance to the omission space
        plambda (float): The parameter calibrated according to the
            human perception of the quality of the super-resolved
            image. Defaults to 0.85.

    Returns:
        torch.Tensor: The relative distance to the center 
        of the improvement space
    """
    H = d_im + d_om -1  
    return d_im + d_om*(1 - np.exp(-H*plambda))

def tc_omission(
    d_im: torch.Tensor,
    d_om: torch.Tensor,
    plambda: float = 0.85
) -> torch.Tensor:
    """ Obtain the relative distance to the center
    of the omission space

    Args:
        d_im (torch.Tensor): The distance to the improvement space
        d_om (torch.Tensor): The distance to the omission space
        plambda (float): The parameter calibrated according to the
            human perception of the quality of the super-resolved
            image. Defaults to 0.85.

    Returns:
        torch.Tensor: The relative distance to the center 
        of the improvement space
    """
    H = d_im + d_om -1  
    return d_om + d_im*(1 - np.exp(-H*plambda))


def tc_hallucination(
    d_im: torch.Tensor,
    d_om: torch.Tensor,
    plambda: float = 0.85
) -> torch.Tensor:
    """ Obtain the relative distance to the center
    of the omission space

    Args:
        d_im (torch.Tensor): The distance to the improvement space
        d_om (torch.Tensor): The distance to the omission space
        plambda (float): The parameter calibrated according to the
            human perception of the quality of the super-resolved
            image. Defaults to 0.85.

    Returns:
        torch.Tensor: The relative distance to the center 
        of the improvement space
    """
    Q = np.exp(-d_im * d_om * plambda)
    return Q
